#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  5 17:30:30 2024

@author: psakic

Low-level functions to import GAMIT metadata file as pandas DataFrame

"""

import datetime as dt
import re

import numpy as np
import pandas as pd

import rinexmod.api as rimo_api

import rinexmod.logger as rimo_log
logger = rimo_log.logger_define("INFO")


def read_gamit_apr_lfile(aprfile_inp):
    """
    read a GAMIT's apr/L-File (GNSS stations coordinates and DOMES)
    and store the data in a DataFrame

    Parameters
    ----------
    aprfile_inp : str
        path of the input apr/L-File

    Returns
    -------
    df : pandas DataFrame
        the apr/lfile info data

    Note
    ----
    If no DOMES is provided in the last 'Note' column,
    a dummy DOMES 00000X000 is returned

    site codes are given in lower case

    """
    lines_stk = []

    for l in open(aprfile_inp):

        if not l[0] == " " or l.strip().startswith("#") or l.strip().startswith("*"):
            continue

        # x, y, z = np.nan, np.nan, np.nan
        # vx,vy,vz = 0.,0.,0.
        # t = pd.NaT

        f = l[1:].split()

        site = l[1:5]
        try:
            site_full = f[0]
        except Exception as e:
            logger.error("unable to read the GAMIT's lfile")
            raise e

        x = float(f[1])
        y = float(f[2])
        z = float(f[3])
        if len(f) >= 8:
            ttmp = float(f[7])
        else:
            ttmp = 1980.0

        # if np.isclose(ttmp , 0.):
        #     t = conv.year_decimal2dt(2000.)
        # else:
        #     t = conv.year_decimal2dt(ttmp)
        t = ttmp

        # if len(l) > 225: # if velocities are given (125 is arbitrary)
        #     vX = float(f[8])
        #     vY = float(f[9])
        #     vZ = float(f[10])
        # else: # if no velocityis given
        #     vX = 0.
        #     vY = 0.
        #     vZ = 0.

        domes_re = re.search("[0-9]{5}[A-Z][0-9]{3}", l)
        if domes_re:
            domes = domes_re.group()
        else:
            domes = "00000X000"

        lines_stk.append((site, site_full, t, x, y, z, domes))

    df = pd.DataFrame(lines_stk)
    df.columns = ["site", "site_full", "epoch", "x", "y", "z", "domes"]

    df["site"] = df["site"].str.lower()

    return df


def read_gamit_station_info(station_info_inp, sort=True):
    """
    read a GAMIT's station.info (GNSS stations metadata)
    and store the data in a DataFrame

    Odd rows in station.info will be skipped

    Parameters
    ----------
    station_info_inp : str
        path of the input station.info.
    sort : bool, optional
        sort the DataFrame by site and start time.
        The default is True.

    Returns
    -------
    df : pandas DataFrame
        the station info data

    Note
    ----
    site codes are given in lower case
    """

    ### from gamit/lib/rstnfo.f
    #
    # c Label         Token  Format   Default      Description
    # c ------        -----  -----    ----------  ----------------------------------------------
    # c *SITE         sitcod  a4                  4-char station code
    # c Station Name  sname   a16                 16-char station name
    # c Session Start  start  2i4,3i3  0 0 0 0 0  start time (yr doy hr min sec)
    # c Session Stop  stop    2i4,3i3  0 0 0 0 0  stop time (yr doy hr min sec)
    # c Sessn         sessn   i1       0          session number
    # c Ant Ht        anth    f7.4     0.0        height of antenna ARP above the monument
    # c Ant N         antn    f7.4     0.0        N offset of ant ARP from the center of the monument
    # c Ant E         ante    f7.4     0.0        E offset of ant ARP from the center of the monument
    # c RcvCod        rcvcod  a6                  6-char/acter GAMIT receiver code
    # c Receiver Type rctype  a20                 IGS (RINEX) receiver name
    # c Receiver SN   rcvrsn  a20                 receiver serial number
    # c SwVer         swver   f5.2     0.0        GAMIT firmware code (real)
    # c Vers          rcvers  a20                 receiver firmware version from RINEX
    # c AntCod        antcod   a6                 6-charr GAMIT antenna code
    # c Antenna SN    antsn   a20                 20-char antenna serial number
    # c Antenna Type  anttyp  a15                 IGS antenna name, 1st 15 chars of RINEX antenna field
    # c Dome          radome   a5                 IGS radome name, last 5 chars of RINEX antenna field
    # c HtCod         htcod    a5     DHARP       5-char GAMIT code for type of height measurement
    # c AntDAZ	antdaz  f5.0     0.0        Alignment from True N (deg).  TAH 2020203.

    # colsize = np.array([4,16,4,4,3,3,3,4,4,3,3,3,1,7,7,7,6,20,20,5,20,6,20,15,5,5,5])
    colsize = np.array(
        [
            5,
            20,
            4 + 1,
            3 + 1,
            2 + 1,
            2 + 1,
            2 + 1,  # start
            4 + 1,
            3 + 1,
            2 + 1,
            2 + 1,
            2 + 1,  # end
            7 + 2,
            7 + 2,
            7 + 2,
            6 + 2,  # antenna high
            20 + 2,
            20 + 2,
            5 + 2,  # rec (2 version fields ...)
            20 + 2,  # rec SN
            15 + 2,
            5 + 2,
            20 + 2,
            1,  # Antenna
        ]
    )

    # colsize_use = colsize+1
    colsize_use = colsize
    # colsize = [e+1  for e in colsize]
    # colsize_use = colsize+2

    col = [
        "site",
        "station name",
        "start year",
        "start doy",
        "start hh",
        "start mm",
        "start ss",
        "stop year",
        "stop doy",
        "stop hh",
        "stop mm",
        "stop ss",
        "ant ht",
        "htcod",
        "ant n",
        "ant e",
        # 'rcvcod',
        "receiver type",
        "vers",
        "swver",
        "receiver sn",
        "antenna type",
        "dome",
        "antenna sn",
        "antdaz",
    ]

    bad_lines = []
    with open(station_info_inp, encoding="iso8859_1") as f:
        for il, l in enumerate(f.readlines()):
            if l[0] != " ":
                bad_lines.append(il)

    df = pd.read_fwf(
        station_info_inp,
        skiprows=bad_lines,
        widths=colsize_use,
        encoding="iso8859_1",
        header=None,
    )

    df.columns = col

    ##### clean df
    ### remove empty rows
    try:
        bool_empty_rows = df["site"].apply(len) == 1
    except Exception as e:
        logger.error("unable to read the GAMIT's station.info")
        raise e

    df = df[np.logical_not(bool_empty_rows)]
    ### do a second NaN cleaning, but the previous should have cleaned everything
    df.dropna(inplace=True, how="all")
    df.reset_index(inplace=True, drop=True)

    ##### create datetime start/end columns
    df["start year"] = df["start year"].replace(9999, 2099)
    df["start doy"] = df["start doy"].replace(999, 365)
    df["start hh"] = df["start hh"].replace(99, 00)
    df["start mm"] = df["start mm"].replace(99, 00)
    df["start ss"] = df["start ss"].replace(99, 00)

    df["stop year"] = df["stop year"].replace(9999, 2099)
    df["stop doy"] = df["stop doy"].replace(999, 365)
    df["stop hh"] = df["stop hh"].replace(99, 00)
    df["stop mm"] = df["stop mm"].replace(99, 00)
    df["stop ss"] = df["stop ss"].replace(99, 00)

    df_start = doy2dt(
        df["start year"],
        df["start doy"],
        df["start hh"],
        df["start mm"],
        df["start ss"],
    )

    df["start"] = df_start

    df_end = doy2dt(
        df["stop year"],
        df["stop doy"],
        df["stop hh"],
        df["stop mm"],
        df["stop ss"]
    )

    df["end"] = df_end

    df["site"] = df["site"].str.lower()

    df["start"] = pd.to_datetime(df["start"])
    df["end"] = pd.to_datetime(df["end"])

    if sort:
        df.sort_values(by=["site", "start"], inplace=True)

    return df


def gamit_df2instru_miscmeta(site, stinfo_df_inp, apr_df_inp, force_fake_coords=False):
    """
    read GAMIT files to get the RinexMod internal
    "instrus" and "misc_meta" dictionaries,
    necessary for the MetaData objects

    Here the site codes must remain as close as
    possible from the GAMIT input files
    i.e. 4 char.

    Parameters
    ----------
    site : str
        native site *4 char.* code which will be extracted from the
        station_info and apriori DataFrame.
    stinfo_df_inp : DataFrame
        station.info-like DataFrame.
    apr_df_inp : DataFrame
        lfile-like DataFrame.
    force_fake_coords : bool, optional
        if True, the coordinates will be set to (0,0,0) and
        the domes number to 00000X000. The default is False.

    Returns
    -------
    installations : dict
        "installations" list i.e. list of "instrus" dict.
    mm_dic : dict
        "misc meta" dict.

    """
    #########################################################
    #### INSTRUMENTATION PART
    stinfo_df_site = stinfo_df_inp[stinfo_df_inp["site"] == site]
    installations = []

    for irow, row in stinfo_df_site.iterrows():
        inst_dic = dict()

        ##### dates
        inst_dic["dates"] = [row["start"].to_pydatetime(), row["end"].to_pydatetime()]

        ##### receiver
        rec_dic = dict()
        rec_dic["Receiver Type"] = str(row["receiver type"])
        rec_dic["Satellite System"] = "GPS+GLO+GAL+BDS+QZSS+SBAS"
        rec_dic["Serial Number"] = str(row["receiver sn"])
        rec_dic["Firmware Version"] = str(row["vers"])
        rec_dic["Elevation Cutoff Setting"] = "0"
        rec_dic["Date Installed"] = row["start"].to_pydatetime()
        rec_dic["Date Removed"] = row["end"].to_pydatetime()
        rec_dic["Temperature Stabiliz."] = "none"
        rec_dic["Additional Information"] = "none"

        inst_dic["receiver"] = rec_dic

        ##### antenna
        ant_dic = dict()
        ant_dic["Antenna Type"] = str(row["antenna type"])
        ant_dic["Serial Number"] = str(row["antenna sn"])
        ant_dic["Antenna Reference Point"] = "none"
        ant_dic["Marker->ARP Up Ecc. (m)"] = row["ant ht"]
        ant_dic["Marker->ARP North Ecc(m)"] = row["ant n"]
        ant_dic["Marker->ARP East Ecc(m)"] = row["ant e"]
        ant_dic["Alignment from True N"] = row["antdaz"]
        ant_dic["Antenna Radome Type"] = str(row["dome"])
        ant_dic["Radome Serial Number"] = "none"
        ant_dic["Antenna Cable Type"] = "none"
        ant_dic["Antenna Cable Length"] = "0"
        ant_dic["Date Installed"] = row["start"].to_pydatetime()
        ant_dic["Date Removed"] = row["end"].to_pydatetime()
        ant_dic["Additional Information"] = "none"
        ant_dic["metpack"] = "none"

        inst_dic["antenna"] = ant_dic

        installations.append(inst_dic)

    #########################################################
    #### MISC META PART
    apr_df_site = apr_df_inp[apr_df_inp["site"] == site]
    if len(apr_df_site) == 0 and not force_fake_coords:
        logger.error(
            "no coords in apr/lfile for %s, abort (you can force fake coords with -fc)",
            site,
        )
        raise rimo_api.RinexModError

    elif len(apr_df_site) == 0 and force_fake_coords:
        # logger.warning("no coords in apr/lfile for %s, fake coords at (0°,0°) used",site)
        apr_df_site = pd.Series(
            {"x": 6378137.000, "y": 0, "z": 0, "domes": "00000X000"}
        )
    else:
        apr_df_site = apr_df_site.iloc[-1]
        apr_df_site.squeeze()

    mm_dic = dict()

    mm_dic["IERS DOMES Number"] = apr_df_site["domes"]

    mm_dic["operator"] = "OPERATOR"
    mm_dic["agency"] = "AGENCY"

    mm_dic["X coordinate (m)"] = apr_df_site["x"]
    mm_dic["Y coordinate (m)"] = apr_df_site["y"]
    mm_dic["Z coordinate (m)"] = apr_df_site["z"]

    mm_dic["ID"] = site
    mm_dic["Country"] = "XXX"

    return installations, mm_dic


def doy2dt(year, days, hours=0, minutes=0, seconds=0):
    """
    Time representation conversion

    Day of Year Time => Python's datetime

    Parameters
    ----------
    year, days : float or list/numpy.array of floats.
        year, days of year
    hours, minutes, seconds : float or list/numpy.array of floats, optional
        hours, minutes, seconds

    Returns
    -------
    L : datetime or list/numpy.array of datetime.
        Datetime(s)

    Note
    ----
    extracted from the GeodeZYX toolbox
    https://github.com/IPGP/geodezyx-toolbox
    """
    if not is_iterable(year):
        # All this because Python cant handle int with a starting with 0 (like 08)
        # => SyntaxError: invalid token

        if np.any(np.isnan([year, days, hours, minutes, seconds])):
            logger.error(
                "one input is NaN, abort: %s,%s,%s,%s,%s",
                year,
                days,
                hours,
                minutes,
                seconds,
            )
            raise Exception

        try:
            year = int(float(str(year)))
            days = int(float(str(days)))
            hours = int(float(str(hours)))
            minutes = int(float(str(minutes)))
            seconds = int(float(str(seconds)))
        except Exception as e:
            logger.error(
                "error with conversion of %s,%s,%s,%s,%s",
                year,
                days,
                hours,
                minutes,
                seconds,
            )
            raise e

        tempsecs = seconds + 60 * minutes + 3600 * hours
        # finalsecs = np.floor(tempsecs)
        finalmicrosec = int(np.round(tempsecs * 10 ** 6))

        return (
                dt.datetime(year, 1, 1)
                + dt.timedelta(days - 1)
                + dt.timedelta(microseconds=finalmicrosec)
        )

    else:
        if not is_iterable(hours):
            hours = [0] * len(year)
        if not is_iterable(minutes):
            minutes = [0] * len(year)
        if not is_iterable(seconds):
            seconds = [0] * len(year)

        outlis = []
        typ = get_type_smart(year)
        for y, d, h, m, s in zip(year, days, hours, minutes, seconds):
            outlis.append(doy2dt(y, d, h, m, s))
        return typ(outlis)


def is_iterable(inp, consider_str_as_iterable=False):
    """
    Test if the input is an iterable like a list or a numpy array or not

    Parameters
    ----------
    inp : list, numpy.array, ...

    consider_str_as_iterable : bool
        string are considered as iterable by Python per default
        This boolean will avoid True as return if you test a string

    Returns
    -------
    out : bool
        True if inp is iterable, False either

    Note
    ----
    extracted from the GeodeZYX toolbox
    https://github.com/IPGP/geodezyx-toolbox
    """
    if not consider_str_as_iterable and type(inp) is str:
        return False

    try:
        iter(inp)
    except TypeError:
        out = False
    else:
        out = True
    return out


def get_type_smart(obj_in):
    """
    get type of an object, to convert easily another one to this type
    for instance type(np.array(A)) doesn't return a constructor

    Note
    ----
    extracted from the GeodeZYX toolbox
    https://github.com/IPGP/geodezyx-toolbox
    """
    typ = type(obj_in)
    if typ is np.ndarray:
        return np.array
    else:
        return typ



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 26/06/2025 18:26:58

@author: psakic
"""

import argparse
import multiprocessing as mp
import os
import re
import subprocess
from datetime import datetime
import hatanaka
import pandas as pd

import rinexmod.api as rimo_api
import rinexmod.api.core_fcts as rimo_cor
import rinexmod.api.rinexmod_main as rimo_main

import rinexmod.logger as rimo_log
logger = rimo_log.logger_define("INFO")

# *****************************************************************************
# Upper level rinexmod for a Console run


def rinexmod_cli(
    rinexinput,
    outputfolder,
    sitelog=None,
    modif_kw=dict(),
    marker="",
    country="",
    longname=False,
    shortname=False,
    force_sitelog=False,
    force_rnx_load=False,
    ignore=False,
    ninecharfile=None,
    no_hatanaka=False,
    compression="gz",
    relative="",
    verbose=True,
    alone=False,
    output_logs=None,
    write=False,
    sort=False,
    full_history=False,
    filename_style="basic",
    multi_process=1,
    debug=False,
    station_info=None,
    lfile_apriori=None,
    force_fake_coords=False,
    remove=False,
    keep_rnx_rec=False,
    round_instru_dates=False,
):
    """
    Main function for reading a Rinex list file. It processes the list, and apply
    file name modification, command line based header modification, or sitelog-based
    header modification.

    Optimized for a CLI (in a Terminal usage) but can be used also in a
    stand-alone API mode.

    For a detailed description, check the help of the lower level
    `rinexmod` function or the help of the frontend CLI function in a Terminal
    """

    # If no longname, modif_kw and sitelog, return
    if (
        not sitelog
        and not modif_kw
        and not marker
        and not longname
        and not shortname
        and not station_info
        and not lfile_apriori
    ):
        logger.critical(
            "No action asked, provide at least one of the following args:"
            "--sitelog, --modif_kw, --marker, --longname, --shortname, --station_info, --lfile_apriori"
        )
        return None

    # If force option provided, check if sitelog option too, if not, not relevant.
    if force_sitelog and not sitelog:
        logger.critical(
            "--force option is relevant only when --sitelog option with a **single** sitelog is also provided"
        )
        return None

    # If ignore option provided, check if sitelog option too, if not, not relevant.
    if ignore and not sitelog:
        logger.critical(
            "--ignore option is relevant only when using also --sitelog option"
        )
        return None

    if ninecharfile and not longname:
        logger.critical(
            "--ninecharfile option is relevant only when using also --longname option"
        )
        return None

    if (station_info and not lfile_apriori) or (not station_info and lfile_apriori):
        logger.critical("--station_info and --lfile_apriori must be provided together")
        return None

    if station_info and lfile_apriori and sitelog:
        logger.critical(
            "both sitelogs and GAMIT files given as metadata input. Managing both is not implemented yet"
        )
        return None

    # If inputfile doesn't exists, return
    if len(rinexinput) == 1 and not os.path.isfile(rinexinput[0]):
        logger.critical("The input file doesn't exist: %s", rinexinput)
        return None

    if len(rinexinput) > 1 and alone:
        logger.critical("several inputs are given while -a/--alone option is set")
        return None

    if output_logs and not os.path.isdir(output_logs):
        logger.critical(
            "The specified output folder for logs doesn't exist : " + output_logs
        )
        return None

    outputfolder = os.path.abspath(outputfolder)
    if not os.path.isdir(outputfolder):
        # mkdirs ???
        os.makedirs(outputfolder)

    # Creating log file
    now = datetime.now()
    nowstr = datetime.strftime(now, "%Y%m%d%H%M%S")

    if output_logs:
        logfolder = output_logs
    else:
        logfolder = outputfolder

    logfile = os.path.join(logfolder, nowstr + "_" + "rinexmod_errors.log")
    if verbose:
        _ = rimo_log.logger_define("DEBUG", logfile, "DEBUG")
    else:
        _ = rimo_log.logger_define("INFO", logfile, "INFO")

    ### the refactor of 2024-03 make this obsolete
    ## now rinexinput is ALWAYS a list
    # if isinstance(rinexinput, list):
    #     pass
    # elif alone:
    #     rinexinput = [rinexinput]
    ### the refactor of 2024-03 make this obsolete

    # Opening and reading lines of the file containing list of rinex to proceed
    if len(rinexinput) == 1 and not alone:
        try:
            rinexinput_use = [line.strip() for line in open(rinexinput[0]).readlines()]
        except:
            errmsg = f"Something went wrong while reading: {str(rinexinput[0])}"
            logger.error(errmsg)
            logger.error("Did you forget -a/--alone option?")
            return rimo_cor.RinexModInputArgsError(errmsg)
    else:
        rinexinput_use = rinexinput

    if not rinexinput_use:
        errmsg = f"The input file is empty: {str(rinexinput)}"
        logger.error(errmsg)
        return rimo_cor.RinexModInputArgsError(errmsg)

    if rinexinput_use[0].endswith("RINEX VERSION / TYPE"):
        errmsg = f"The input file is not a file list but a RINEX: {str(rinexinput[0])}"
        logger.error(errmsg)
        logger.error("Did you forget -a/--alone option?")
        return rimo_cor.RinexModInputArgsError(errmsg)

    # sort the RINEX list
    if sort:
        rinexinput_use.sort()

    ### load the sitelogs/GAMIT-files as a **list of MetaData objects**
    # from sitelogs
    if sitelog:
        sitelogs_list_use = rimo_cor.metadata_input_manage(sitelog, force_sitelog)
    # from GAMIT files
    elif station_info and lfile_apriori:
        sitelogs_list_use = rimo_cor.gamit2mda_objs(
            station_info, lfile_apriori, force_fake_coords=force_fake_coords
        )
    else:
        sitelogs_list_use = None

    ### Looping in file list ###
    return_lists = dict()
    ####### Iterate over each RINEX
    rnxmod_kwargs_lis = []
    for rnx in rinexinput_use:
        rnxmod_kwargs = {
            "rinexfile": rnx,
            "outputfolder": outputfolder,
            "sitelog": sitelogs_list_use,
            "modif_kw": modif_kw,
            "marker": marker,
            "country": country,
            "longname": longname,
            "shortname": shortname,
            "force_rnx_load": force_rnx_load,
            "force_sitelog": force_sitelog,
            "ignore": ignore,
            "ninecharfile": ninecharfile,
            "no_hatanaka": no_hatanaka,
            "compression": compression,
            "relative": relative,
            "verbose": verbose,
            "return_lists": return_lists,
            "full_history": full_history,
            "filename_style": filename_style,
            "station_info": station_info,
            "lfile_apriori": lfile_apriori,
            "force_fake_coords": force_fake_coords,
            "remove": remove,
            "keep_rnx_rec": keep_rnx_rec,
            "round_instru_dates": round_instru_dates,
        }

        rnxmod_kwargs_lis.append(rnxmod_kwargs)

    global rinexmod_mpwrap

    def rinexmod_mpwrap(rnxmod_kwargs_inp):
        try:
            return_lists_out = rimo_main.rinexmod(**rnxmod_kwargs_inp)
            return return_lists_out
        except Exception as e:
            if debug:  ### set as True for debug mode
                raise e
            else:
                logger.error(
                    "%s raised, RINEX is skiped: %s. Use -d/--debug option for more details",
                    type(e).__name__,
                    rnxmod_kwargs_inp["rinexfile"],
                )
                return None

    # number of parallel processing
    if multi_process > 1:
        logger.info("multiprocessing: %d cores used", multi_process)
    pool = mp.Pool(processes=multi_process)
    res_raw = [pool.apply_async(rinexmod_mpwrap, args=(x,)) for x in rnxmod_kwargs_lis]
    results = [e.get() for e in res_raw]

    for return_lists_mono in results:
        try:
            rimo_cor.rtun_lsts_make(return_lists_mono, return_lists)
        except rimo_cor.ReturnListError:
            logger.warning(
                "one file has been skipped because of an odd individual return list"
            )
            continue

    #########################################
    logger.handlers.clear()

    if write:
        rimo_cor.rtun_lsts_write(return_lists, logfolder, now)

    return return_lists


##### Class for --modif_kw
class ParseKwargs(argparse.Action):
    # source:
    # https://sumit-ghosh.com/posts/parsing-dictionary-key-value-pairs-kwargs-argparse-python/
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        icmt = 0
        for value in values:
            try:
                key, value = value.split("=")

                if key == "comment":
                    getattr(namespace, self.dest)[key + "_" + str(icmt)] = value
                    icmt += 1
                else:
                    getattr(namespace, self.dest)[key] = value

            except Exception as e:

                def _print_kw_tips(values_inp):
                    logger.critical("********************************************")
                    logger.critical("TIP1: be sure you have respected the syntax:")
                    logger.critical("      -k keyword1='value1' keyword2='value2'")
                    logger.critical("********************************************")
                    logger.critical(values_inp)
                    return None

                _print_kw_tips(values)
                raise e


# *****************************************************************************



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Class
v1 - 2021-02-07 Félix Léger - felixleger@gmail.com
v2 - 2024-02-26 Pierre Sakic - sakic@ipgp.fr
"""

import configparser
import copy
import json
import os
import re
from datetime import datetime, timedelta
import pycountry

import pandas as pd

import rinexmod.common.gamit_meta as rimo_gmm
import rinexmod.classes as rimo_cls

import rinexmod.api as rimo_api

import rinexmod.logger as rimo_log

logger = rimo_log.logger_define("INFO")


class MetaData:
    """
    Parses and stores in a dict information from an IGS sitelog or GAMIT-like files.
    Requires one parameter, the sitelog path.
    At instantiation, will parse the sitelog and store in a dict all parsed values.
    Dict accessible via MetaData.raw_content
    Will also create a tab, stored in MetaData.instrus, containing all
    the different instrumentation periods, tab containing a start and an end date,
    and for each line a dict of antenna instrumentation an receiver instrumentation.

    New 20240225: MetaData object can also be initialized as an empty one.
    (for GAMIT's station.info/apr/L-File import)
    thus sitelogfile can now be None

    4 available methods:
    * find_instru takes a start and an end date and returns the instrumentation
    corresponding to the period, if found. Option to ignore firmware version inconsistency.
    * teqcargs also takes a start and an end date and returns a string of args to
    pass to teqc so that it will modify a rinex file's header.
    * find_instru4rnx will return a dict with all header metadatas that is
    compatible with RinexFile header modifications methods.
    * write_json will write the dict of the parsed values from the sitelog to a
    json file.
    """

    def __init__(self, sitelogfile=None):
        if sitelogfile:
            ### usual case. read from a sitelog file
            self.set_from_sitelog(sitelogfile)
        else:
            ### more generic and flexible case. empty object
            self.path = None
            self.filename = None
            self._site_id = None
            # self.site4char = None
            # site9char is a more complex property bellow
            self.raw_content = None
            self.instrus = []
            self.misc_meta = {}
            self.raw_content_apr = None

    ### site
    @property
    def site_id(self):
        return self._site_id

    @site_id.setter
    def site_id(self, value):
        self._site_id = value

    @property
    def site_id4(self):
        return self._site_id[:4].lower()

    @property
    def site_id9(self):
        return rimo_api.make_site_id9(self.site_id)

    def __repr__(self):
        return "{} metadata, from {}".format(self.site_id, self.filename)

    def set_misc_meta(
        self, site_id, domes, operator, agency, x, y, z, date_prepared, country
    ):
        """
        Exemple of misc meta dict:

        {
         'ID': 'PMZI00MYT',
         'IERS DOMES Number': '90109M001',
         'operator': 'OVPF-IPGP',
         'agency': 'IPGP',
         'X coordinate (m)': 4377557.000,
         'Y coordinate (m)': 4419689.300,
         'Z coordinate (m)': -1403760.500,
         'date prepared': datetime.datetime(2024, 9, 13, 0, 0),
         'Country': 'Mayotte'
        }
        """

        self.site_id = site_id

        self.misc_meta["ID"] = site_id
        self.misc_meta["IERS DOMES Number"] = domes
        self.misc_meta["operator"] = operator
        self.misc_meta["agency"] = agency
        self.misc_meta["X coordinate (m)"] = float(x)
        self.misc_meta["Y coordinate (m)"] = float(y)
        self.misc_meta["Z coordinate (m)"] = float(z)
        self.misc_meta["date prepared"] = date_prepared
        self.misc_meta["Country"] = country

        return self.misc_meta

    def set_from_sitelog(self, sitelogfile):
        """
        initialization method for metadata import from sitelog
        """
        self.path = sitelogfile
        self.filename = os.path.basename(self.path)
        self.raw_content = self.slg_file2raw()

        if self.raw_content:
            self.instrus = self.slg_raw2instrus()
            self.misc_meta = self.slg_raw2misc_meta()
            self.site_id = self.misc_meta["ID"]

        else:
            self.instrus = None
            self.misc_meta = None
            self.site_id = self.filename[:4].lower()

        return None

    def set_from_gamit(
        self,
        site,
        station_info,
        lfile,
        force_fake_coords=False,
        station_info_name="station.info",
    ):
        """
        initialization method for metadata import from GAMIT files

        site can be 4 char or 9 char
        """

        ##### This site_id management is very strange, must be clarified
        self.site_id = site[:4].lower()
        if len(site) == 9:
            self.misc_meta["ID"] = site.upper()
            self.misc_meta["Country"] = site[-3:].upper()

        if isinstance(station_info, pd.DataFrame):
            self.raw_content = station_info
            self.path = None
            self.filename = station_info_name
        else:
            self.raw_content = rimo_gmm.read_gamit_station_info(self.path)
            self.path = station_info
            self.filename = os.path.basename(self.path)

        if isinstance(lfile, pd.DataFrame):
            self.raw_content_apr = lfile
        else:
            self.raw_content_apr = rimo_gmm.read_gamit_apr_lfile(lfile)

        if self.raw_content is not None:
            self.instrus, self.misc_meta = rimo_gmm.gamit_df2instru_miscmeta(
                site=self.site_id4,
                stinfo_df_inp=self.raw_content,
                apr_df_inp=self.raw_content_apr,
                force_fake_coords=force_fake_coords,
            )
        else:
            self.instrus, self.misc_meta = None, None

        return None

    def set_from_rinex(self, rnxfile):
        """
        initialization method for metadata import from rinex header
        """
        self.path = rnxfile
        self.filename = os.path.basename(self.path)

        rnxobj = rimo_cls.RinexFile(self.path)
        head_str, head_dic = rnxobj.get_header()

        self.site_id = head_dic["Marker name"]

        self.raw_content = head_str

        inst_dic = dict()
        ##### receiver
        rec_dic = dict()
        rec_dic["Receiver Type"] = head_dic["Receiver type"]
        rec_dic["Satellite System"] = "GPS+GLO+GAL+BDS+QZSS+SBAS"
        rec_dic["Serial Number"] = head_dic["Receiver serial"]
        rec_dic["Firmware Version"] = head_dic["Receiver firmware version"]
        rec_dic["Elevation Cutoff Setting"] = "0"
        rec_dic["Date Installed"] = head_dic["Start date and time"]
        rec_dic["Date Removed"] = head_dic["Final date and time"]
        rec_dic["Temperature Stabiliz."] = "none"
        rec_dic["Additional Information"] = "none"

        inst_dic["receiver"] = rec_dic

        ##### antenna
        ant_dic = dict()
        ecc = head_dic["Antenna delta (H/E/N)"].split()
        ant_dic["Antenna Type"] = head_dic["Antenna type"].strip()[:-4]
        ant_dic["Serial Number"] = head_dic["Antenna serial"]
        ant_dic["Antenna Reference Point"] = "none"
        ant_dic["Marker->ARP Up Ecc. (m)"] = float(ecc[0])
        ant_dic["Marker->ARP North Ecc(m)"] = float(ecc[1])
        ant_dic["Marker->ARP East Ecc(m)"] = float(ecc[2])
        ant_dic["Alignment from True N"] = 0.0
        ant_dic["Antenna Radome Type"] = head_dic["Antenna type"].strip()[-4:]
        ant_dic["Radome Serial Number"] = "none"
        ant_dic["Antenna Cable Type"] = "none"
        ant_dic["Antenna Cable Length"] = "0"
        ant_dic["Date Installed"] = head_dic["Start date and time"]
        ant_dic["Date Removed"] = head_dic["Final date and time"]
        ant_dic["Additional Information"] = "none"
        ant_dic["metpack"] = "none"

        inst_dic["antenna"] = ant_dic

        inst_dic["dates"] = [
            head_dic["Start date and time"],
            head_dic["Final date and time"],
        ]

        self.instrus = [inst_dic]

        xyz = head_dic["Antenna position (XYZ)"].split()
        self.set_misc_meta(
            site_id=head_dic["Marker name"],
            domes=head_dic["Marker number"],
            operator=head_dic["Operator"],
            agency=head_dic["Agency"],
            x=float(xyz[0]),
            y=float(xyz[1]),
            z=float(xyz[2]),
            date_prepared=head_dic["Start date and time"],
            country="XXX",
        )

        return None

    def add_instru(self, rec_dic: dict, ant_dic: dict, date_srt=None, date_end=None):
        """
        Add an instrumentation period to instrus attribute the metadata object.

        Example of an instru dict:

         {
        'dates': [datetime.datetime(2008, 7, 8, 4, 48),
                  datetime.datetime(2009, 1, 1, 0, 0)],

        'receiver': {'Receiver Type': 'TPS GB-1000',
                     'Satellite System': 'GPS+GLO',
                     'Serial Number': 'T225373',
                     'Firmware Version': '3.32',
                     'Elevation Cutoff Setting': '15 deg',
                     'Date Installed': datetime.datetime(2008, 7, 8, 4, 48),
                     'Date Removed': datetime.datetime(2010, 3, 16, 5, 0),
                     'Temperature Stabiliz.': 'none',
                     'Additional Information': '(multiple lines)'},

          'antenna': {'Antenna Type': 'ASH701975.01A   NONE',
                      'Serial Number': '8279',
                      'Antenna Reference Point': 'TOP',
                      'Marker->ARP Up Ecc. (m)': '0.0000',
                      'Marker->ARP North Ecc(m)': '0.0000',
                      'Marker->ARP East Ecc(m)': '0.0000',
                      'Alignment from True N': '0 deg',
                      'Antenna Radome Type': 'NONE',
                      'Radome Serial Number': '',
                      'Antenna Cable Type': 'TNC',
                      'Antenna Cable Length': '4 m',
                      'Date Installed': datetime.datetime(2008, 3, 14, 0, 0),
                      'Date Removed': datetime.datetime(2009, 1, 1, 0, 0),
                      'Additional Information': 'La date de desinstallation est inconnue'},
                      'metpack': None
          }
        """

        date_srt = date_srt or datetime(1980, 1, 1)
        date_end = date_end or datetime(2099, 1, 1)

        instru_dic = dict()

        ### dates
        instru_dic["dates"] = [date_srt, date_end]
        ### receiver
        instru_dic["receiver"] = rec_dic
        instru_dic["receiver"]["Date Installed"] = date_srt
        instru_dic["receiver"]["Date Removed"] = date_end
        ### antenna
        instru_dic["antenna"] = ant_dic
        if (
            "Antenna Type" in ant_dic.keys()
            and not "Antenna Radome Type" in ant_dic.keys()
        ):
            ant_dic["Antenna Radome Type"] = ant_dic["Antenna Radome Type"][-4:]
        instru_dic["antenna"]["Date Installed"] = date_srt
        instru_dic["antenna"]["Date Removed"] = date_end
        ### append to instrus
        self.instrus.append(instru_dic)
        return instru_dic

    def sort_instrus(self):
        """
        Sort the instrumentation periods in the instrus attribute
        """

        self.instrus = sorted(self.instrus, key=lambda x: x["dates"][0])

        return self.instrus

    def merge_instrus(self):
        """
        Merge overlapping instrumentation periods in the `instrus` attribute.

        This method processes the `instrus` attribute, which is a list of instrumentation
        periods, and merges periods that overlap or are considered equal based on the
        `equal_instru` function. The merged periods are stored in the `instrus` attribute.

        Returns
        -------
        list
            A list of merged instrumentation periods.
        """

        instrus_out = []
        # Sort the instrumentation periods
        self.sort_instrus()
        instru_use = copy.deepcopy(self.instrus[0])

        for ins in self.instrus[1:]:
            # If the dates are overlapping, we merge the two periods
            if equal_instru(instru_use, ins):
                # Merge the two periods
                instru_use["dates"][0] = min(instru_use["dates"][0], ins["dates"][0])
                instru_use["dates"][1] = max(instru_use["dates"][1], ins["dates"][1])

                # Merge the receiver and antenna dictionaries
                for key in ["receiver", "antenna"]:
                    instru_use[key].update(ins[key])
            else:
                instrus_out.append(instru_use)
                instru_use = copy.deepcopy(ins)

        instrus_out.append(instru_use)

        self.instrus = instrus_out
        self.instrus_dates2recant_install()

        return self.instrus

    def instrus_recant_install2dates(self, date_source="receiver"):
        """
        Update the `dates` key in each instrumentation period based on the installation
        and removal dates of either the receiver or the antenna.

        This method processes the `instrus` attribute, which is a list of instrumentation
        periods, and updates the `dates` key for each period using the installation and
        removal dates from either the receiver or the antenna, depending on the specified
        `date_source`.

        Parameters
        ----------
        date_source : str, optional
            Specifies whether to use the receiver or antenna dates to update the `dates` key.
            Must be either "receiver" or "antenna". Default is "receiver".

        Returns
        -------
        list
            The updated list of instrumentation periods with the `dates` key set based on
            the specified `date_source`.

        Raises
        ------
        ValueError
            If `date_source` is not "receiver" or "antenna".
        """
        for ins in self.instrus:
            if date_source == "receiver":
                ins["dates"] = [
                    ins["receiver"]["Date Installed"],
                    ins["receiver"]["Date Removed"],
                ]
            elif date_source == "antenna":
                ins["dates"] = [
                    ins["antenna"]["Date Installed"],
                    ins["antenna"]["Date Removed"],
                ]
            else:
                raise ValueError("date_source must be 'receiver' or 'antenna'")
        return self.instrus

    def instrus_dates2recant_install(self):
        """
        Update the receiver and antenna installation dates in the `instrus` attribute
        based on the dates stored in the `dates` key of each instrumentation period.

        This method processes the `instrus` attribute, which is a list of instrumentation
        periods, and updates the installation dates for both the receiver and antenna
        based on the dates stored in the `dates` key.

        Returns
        -------
        list
            A list of instrumentation periods with updated installation dates.
        """

        for ins in self.instrus:
            ins["receiver"]["Date Installed"] = ins["dates"][0]
            ins["receiver"]["Date Removed"] = ins["dates"][1]
            ins["antenna"]["Date Installed"] = ins["dates"][0]
            ins["antenna"]["Date Removed"] = ins["dates"][1]

        return self.instrus

    #  _____               _                __                  _   _
    # |  __ \             (_)              / _|                | | (_)
    # | |__) |_ _ _ __ ___ _ _ __   __ _  | |_ _   _ _ __   ___| |_ _  ___  _ __  ___
    # |  ___/ _` | '__/ __| | '_ \ / _` | |  _| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
    # | |  | (_| | |  \__ \ | | | | (_| | | | | |_| | | | | (__| |_| | (_) | | | \__ \
    # |_|   \__,_|_|  |___/_|_| |_|\__, | |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
    #                               __/ |
    #                              |___/

    def slg_file2raw(self, keys_float=False):
        """
        First function for reading a Sitelog file.
        From the sitelog file,
        returns a dict with all readed values.
        """
        ###### Input and output file tests #######

        # Checking if inexisting file
        if not os.path.isfile(self.path):
            # print('The provided Sitelog is not valid : ' + self.path)
            return None, 2

        # Getting filename and basename for output purposes
        # filename = (os.path.splitext(os.path.basename(self.path))[0])
        # dirname = os.path.dirname(self.path)

        ####### Reading Sitelog File #########

        # Reading the sitelog file
        try:
            with open(self.path, "r", encoding="utf-8") as datafile:
                sitelog = datafile.read()
        except UnicodeDecodeError:  # OVPF sitelogs are in iso-8859-1
            try:
                with open(self.path, "r", encoding="iso-8859-1") as datafile:
                    sitelog = datafile.read()
            except:
                raise

        # We delete all initial space.
        pattern = re.compile(r"\n +")
        sitelog = re.sub(pattern, r"\n", sitelog)

        # We rearrange multiline content to be complient with .ini format.
        pattern = re.compile(r"(\n *): *")
        sitelog = re.sub(pattern, " ", sitelog)

        # We transform  multiple contacts into sub blocs
        pattern = re.compile(r"((?:Secondary|Primary) [Cc]ontact):{0,1}")
        sitelog = re.sub(pattern, r"[\1]", sitelog)

        # We remove the final graphic if exists
        antennagraphic = re.search(r"Antenna Graphics with Dimensions", sitelog)
        if antennagraphic:
            sitelog = sitelog[: antennagraphic.start(0)]

        # List of formated blocs
        formatedblocs = []
        # Final dict to store values
        slgdic = {}

        # We split the file into major blocs (reading the '4.'' type pattern)
        itr = re.finditer(r"\d{1,2}\. +.+\n", sitelog)
        indices = [m.start(0) for m in itr]

        blocs = [sitelog[i:j] for i, j in zip(indices, indices[1:] + [None])]

        if len(blocs) == 0:
            # print('The provided Sitelog is not correct : ' + self.path)
            return None

        # We loop into those blocs, after a test that permits keeping only blocs
        # beginning with patterns like '6.'. This permits removing the title bloc.
        for bloc in [bloc for bloc in blocs if re.match(r"\d.", bloc[:2])]:

            # We search for '4.3', '4.3.', '4.2.3' patterns for subbloc detection
            itr = re.finditer(r"\n\d{1,2}\.\d{0,2}\.{0,1}\w{0,2}\.{0,1}", bloc)
            indices = [m.start(0) + 1 for m in itr]

            if len(indices) > 0:  # If subblocs
                subblocs = [bloc[i:j] for i, j in zip(indices, indices[1:] + [None])]

                for subbloc in subblocs:
                    # We separate index (the first line) from values
                    index, subbloc = subbloc.split("\n", 1)
                    # If available, the data contained in the first line (now stored in index)
                    # is pushed back in the subbloc varaible in a new 'type' entry.
                    try:
                        index, title = index.split(" ", 1)
                        if ":" not in title:
                            title = "type : " + title
                        subbloc = title.lstrip() + "\n" + subbloc
                    except Exception:
                        pass
                    # We append the subbloc to the list of blocs to read
                    formatedblocs.append([index, subbloc])

            elif re.search(r"\n", bloc):
                # Get index and bloc content
                index, bloc = bloc.split("\n", 1)
                index = re.match(r"\d{1,2}\.", index).group(0)

                # We append it to the list of blocs to read
                formatedblocs.append([index, bloc])

            else:
                pass

        # Now that blocs are formated, we read them with configparser
        for [index, bloc] in formatedblocs:

            if "x" in index[0:5]:
                pass  # If it's a model section (like 3.x), we don't proceed it
            else:
                # We add a section header to work on it with ConfigParser
                bloc = "[dummy_section]\n" + bloc

                cfgparser = configparser.RawConfigParser(allow_no_value=True)
                cfgparser.optionxform = str  # Respect case
                cfgparser.read_string(bloc)

                # We construct the bloc dict
                blocdict = {}
                for section_name in cfgparser.sections():
                    # For 'dummy_section' section, we quit the section_name
                    if section_name == "dummy_section":
                        blocdict.update(dict(cfgparser[section_name]))
                    # For other sections (Primary & Secondary contact, added earlier), we keep the section_name
                    else:
                        blocdict.update({section_name: dict(cfgparser[section_name])})

                # We append the bloc dict to the global dict
                if keys_float:
                    keys_contact = [11.0, 12.0]
                    slgdic[float(index)] = blocdict
                else:
                    keys_contact = ["11.", "12."]
                    slgdic[index] = blocdict

        # Contact corrections - putting the field 'Additional Information' in the right level dict
        # and removing network information
        for key in [key for key in slgdic.keys() if key in keys_contact]:
            if "network" in slgdic[key]["Agency"].lower():
                index_network = slgdic[key]["Agency"].lower().index("network")
                slgdic[key]["Agency"] = slgdic[key]["Agency"][:index_network]
            # Removing extra spaces
            slgdic[key]["Agency"] = slgdic[key]["Agency"].strip()
            slgdic[key]["Agency"] = " ".join(slgdic[key]["Agency"].split())
            if slgdic[key]["Secondary Contact"]["Additional Information"]:
                # Putting the 'Additional Information' in the lower level dict
                slgdic[key]["Additional Information"] = slgdic[key][
                    "Secondary Contact"
                ]["Additional Information"]
                # Removing it from the incorrect dict level
                slgdic[key]["Secondary Contact"].pop("Additional Information", None)

        return slgdic

    def slg_raw2instrus(self):
        """
        This function identifies the different complete installations from the
        antenna and receiver change dates, then returns a list of dictionnaries
         with only instrumented periods.

        It uses the raw_content attribute (dictionary), generated with
        gen_raw_content_dict method.

        The output is a list containing one or several dictionaries with 3 keys
        'dates' 'receiver' 'antenna' and the following structure:

        Example of an instru dict:

         {
        'dates': [datetime.datetime(2008, 7, 8, 4, 48),
                  datetime.datetime(2009, 1, 1, 0, 0)],

        'receiver': {'Receiver Type': 'TPS GB-1000',
                     'Satellite System': 'GPS+GLO',
                     'Serial Number': 'T225373',
                     'Firmware Version': '3.32',
                     'Elevation Cutoff Setting': '15 deg',
                     'Date Installed': datetime.datetime(2008, 7, 8, 4, 48),
                     'Date Removed': datetime.datetime(2010, 3, 16, 5, 0),
                     'Temperature Stabiliz.': 'none',
                     'Additional Information': '(multiple lines)'},

          'antenna': {'Antenna Type': 'ASH701975.01A   NONE',
                      'Serial Number': '8279',
                      'Antenna Reference Point': 'TOP',
                      'Marker->ARP Up Ecc. (m)': '0.0000',
                      'Marker->ARP North Ecc(m)': '0.0000',
                      'Marker->ARP East Ecc(m)': '0.0000',
                      'Alignment from True N': '0 deg',
                      'Antenna Radome Type': 'NONE',
                      'Radome Serial Number': '',
                      'Antenna Cable Type': 'TNC',
                      'Antenna Cable Length': '4 m',
                      'Date Installed': datetime.datetime(2008, 3, 14, 0, 0),
                      'Date Removed': datetime.datetime(2009, 1, 1, 0, 0),
                      'Additional Information': 'La date de desinstallation est inconnue'},
                      'metpack': None
          }

        Returns
        -------
        instrus : list
        """

        ##### Constructing a list of date intervals from all changes dates #####

        listdates = []

        # We extract dates for blocs 3. and 4. (reveiver, antenna)
        keys = [
            k
            for k in self.raw_content.keys()
            if k.startswith("3.") or k.startswith("4.")
        ]

        for key in keys:
            # Formating parsed dates - set empty to 'infinity' date. If not a date, it's because it's an open border.
            self.raw_content[key]["Date Installed"] = self._tryparsedate(
                self.raw_content[key]["Date Installed"]
            )
            self.raw_content[key]["Date Removed"] = self._tryparsedate(
                self.raw_content[key]["Date Removed"]
            )
            # Adding dates to listdate
            listdates += (
                self.raw_content[key]["Date Installed"],
                self.raw_content[key]["Date Removed"],
            )

        # Quitting null values
        listdates = [date for date in listdates if date]
        # Quitting duplicates
        listdates = list(set(listdates))
        # Sorting
        listdates.sort()

        # List of installations. An installation is a date interval, a receiver and an antena
        instrus = []

        # Constructing the installations list - date intervals
        for i in range(0, len(listdates) - 1):
            # Construct interval from listdates
            dates = [listdates[i], listdates[i + 1]]
            # Setting date interval in Dict of installation
            ins = dict(dates=dates, receiver=None, antenna=None, metpack=None)
            # Append it to list of installations
            instrus.append(ins)

        ##### Getting Receiver info for each interval #####

        receivers = [
            self.raw_content[key]
            for key in self.raw_content.keys()
            if key.startswith("3.")
        ]

        # Constructing the installations list - Receivers
        for ins in instrus:
            # We get the receiver corresponding to the date interval
            for receiver in receivers:
                if (receiver["Date Installed"] <= ins["dates"][0]) and (
                    receiver["Date Removed"] >= ins["dates"][1]
                ):
                    ins["receiver"] = receiver
                    # Once found, we quit the loop
                    break

        ##### Getting Antenna info for each interval #####

        antennas = [
            self.raw_content[key]
            for key in self.raw_content.keys()
            if key.startswith("4.")
        ]

        # Constructing the installations list - Antennas
        for ins in instrus:
            # We get the antenna corresponding to the date interval
            for antenna in antennas:
                if (antenna["Date Installed"] <= ins["dates"][0]) and (
                    antenna["Date Removed"] >= ins["dates"][1]
                ):
                    ins["antenna"] = antenna
                    # Once found, we quit the loop
                    break

        ##### Removing from installation list periods without antenna or receiver

        instrus = [i for i in instrus if i["receiver"] and i["antenna"]]

        return instrus

    @staticmethod
    def _tryparsedate(date):
        # Different date format to test on the string in case of bad standard compliance
        formats = [
            "%Y-%m-%d %H:%M:%S.%f",
            "%Y-%m-%dT%H:%MZ",
            "%Y-%m-%d %H:%M",
            "%Y-%m-%dT%H:%M",
            "%Y/%m/%dT%H:%MZ",
            "%Y-%m-%d %H:%M",
            "%Y-%m-%dT%H:%M",
            "%d/%m/%YT%H:%MZ",
            "%d/%m/%Y %H:%M",
            "%d/%m/%YT%H:%M",
            "%Y-%m-%d",
            "%Y/%m/%d",
            "%d/%m/%Y",
        ]
        if date:
            # Parse to date trying different formats
            for fmt in formats:
                try:
                    date = datetime.strptime(date, fmt)
                    break
                except:
                    pass
        if not isinstance(date, datetime):
            # We set the date to 'infinity' date. If not a date, it's because it's an open border.
            date = datetime.strptime("9999-01-01", "%Y-%m-%d")

        return date

    def slg_raw2misc_meta(self):
        """
        This function generates the "misc meta" dictionary, i.e. a
        dictionary containing all the useful metadata information which are not
        stored in the instrumentation dictionary
        (see slg_raw2instrus )

        Consistent with [IGSMAIL-8458] (2024-06-01)

        Exemple of misc meta dict:
        {
         'ID': 'PMZI00MYT',
         'IERS DOMES Number': '90109M001',
         'operator': 'OVPF-IPGP',
         'agency': 'IPGP',
         'X coordinate (m)': '4377557.000',
         'Y coordinate (m)': '4419689.300',
         'Z coordinate (m)': '-1403760.500',
         'date prepared': datetime.datetime(2024, 9, 13, 0, 0),
         'Country': 'Mayotte'
        }
        """

        if (
            "Nine Character ID" in self.raw_content["1."].keys()
        ):  # now consistent with [IGSMAIL-8458]
            site_id = self.raw_content["1."]["Nine Character ID"]
        else:
            site_id = self.raw_content["1."]["Four Character ID"]

        domes = self.raw_content["1."]["IERS DOMES Number"]

        operator = self.raw_content["11."]["Preferred Abbreviation"]
        agency = self.raw_content["12."]["Preferred Abbreviation"]

        x = self.raw_content["2."]["X coordinate (m)"]
        y = self.raw_content["2."]["Y coordinate (m)"]
        z = self.raw_content["2."]["Z coordinate (m)"]

        date_prepared = datetime.strptime(
            self.raw_content["0."]["Date Prepared"], "%Y-%m-%d"
        )

        if (
            "Country/Region" in self.raw_content["2."].keys()
        ):  # now consistent with [IGSMAIL-8458]
            country = self.raw_content["2."]["Country/Region"]
        elif (
            "Country or Region" in self.raw_content["2."].keys()
        ):  # now consistent with [IGSMAIL-8458]
            country = self.raw_content["2."]["Country or Region"]
        else:
            country = self.raw_content["2."]["Country"]

        self.misc_meta = dict()
        # We must initialize the misc_meta here
        # not initialized before (we are in the sitelog case)
        mm_dic = self.set_misc_meta(
            site_id=site_id,
            domes=domes,
            operator=operator,
            agency=agency,
            x=x,
            y=y,
            z=z,
            date_prepared=date_prepared,
            country=country,
        )

        return mm_dic

    #  ______                         _   _   _                __                  _   _
    # |  ____|                       | | | | (_)              / _|                | | (_)
    # | |__ ___  _ __ _ __ ___   __ _| |_| |_ _ _ __   __ _  | |_ _   _ _ __   ___| |_ _  ___  _ __  ___
    # |  __/ _ \| '__| '_ ` _ \ / _` | __| __| | '_ \ / _` | |  _| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
    # | | | (_) | |  | | | | | | (_| | |_| |_| | | | | (_| | | | | |_| | | | | (__| |_| | (_) | | | \__ \
    # |_|  \___/|_|  |_| |_| |_|\__,_|\__|\__|_|_| |_|\__, | |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
    #                                                  __/ |
    #                                                 |___/

    def find_instru(self, starttime, endtime, ignore=False, round_instru_dates=False):
        """
        Get the instrumentation corresponding to starttime and endtime.
        If ignore=True, firmware version changes between periods are ignored.
        If round_instru_dates=True, instru dates are rounded to day boundaries.
        """

        # Try exact match first

        dt1d = timedelta(minutes=1440)

        for ins in self.instrus:

            if round_instru_dates:
                ins_srt = rimo_api.round_time(
                    ins["dates"][0], date_delta=dt1d, to="down"
                )
                ins_end = rimo_api.round_time(
                    ins["dates"][1], date_delta=dt1d, to="up"
                )
            else:
                ins_srt = ins["dates"][0]
                ins_end = ins["dates"][1]

            if ins_srt <= starttime and ins_end >= endtime:
                return ins, False

        # If not found and ignore=True, check consecutive periods ignoring firmware
        if not ignore:
            return None, False

        # BACKUP CASE
        # If we can't find a corresponding installation period and we use the force
        # option, we will force ignoring the firmware version modification and
        # consider only the other parameters
        def compare_instru_ignore_firmware(instru1, instru2):
            """Helper to compare two instru periods ignoring dates and firmware."""
            exclude_keys = [
                "Date Removed",
                "Date Installed",
                "Additional Information",
                "Firmware Version",
            ]

            for key in ["antenna", "receiver"]:
                dict1 = {k: v for k, v in instru1[key].items() if k not in exclude_keys}
                dict2 = {k: v for k, v in instru2[key].items() if k not in exclude_keys}
                if dict1 != dict2:
                    return False

            return True

        for i in range(len(self.instrus) - 1):
            if not (
                self.instrus[i]["dates"][0] <= starttime
                and self.instrus[i + 1]["dates"][1] >= endtime
            ):
                continue

            # Compare periods without dates, firmware, and additional info
            if compare_instru_ignore_firmware(self.instrus[i], self.instrus[i + 1]):
                return self.instrus[i], True

        return None, False

    def get_country(self, iso_code=True):
        """
        Return the ISO country code based on the Sitelog's Country field.
        Requires pycountry module

        Consistent with [IGSMAIL-8458] (2024-06-01)
        """

        raw_country = self.misc_meta["Country"]

        #### The raw input is a 3-letter ISO code, now consistent with [IGSMAIL-8458]
        if len(raw_country) == 3:
            iso_country = raw_country
            try:
                full_country = pycountry.countries.get(alpha_3=iso_country).name
            except:
                full_country = "Somewhere"

        #### The raw input is a full country name
        else:
            full_country = raw_country
            full_country2 = full_country.split("(the)")[0].strip()
            try:
                iso_country = pycountry.countries.get(name=full_country2).alpha_3
            except:
                iso_country = "XXX"

        if iso_code:
            return iso_country
        else:
            return full_country

    def teqcargs(self, starttime, endtime, ignore=False):
        """
        Will return a string of teqc args containing all infos from the sitelog,
        including instrumentation infos taking into account a start and an end date.
        If ignore option set to True, we will force ignoring the firmware version
        modification between two periods and consider only the other parameters.
        """

        instru, ignored = self.find_instru(starttime, endtime, ignore)

        if not instru:
            return "", ignored

        ########### GNSS one-letter codes ###########

        # From https://www.unavco.org/software/data-processing/teqc/tutorial/tutorial.html
        gnss_codes = dict(
            GPS="G", GLO="R", GAL="E", BDS="C", QZSS="J", IRNSS="I", SBAS="S"
        )

        # GNSS system. M if multiple, else, one letter code from gnss_codes dict.
        o_system = instru["receiver"]["Satellite System"]
        if "+" in o_system:
            o_system = "M"
        else:
            o_system = gnss_codes[o_system]

        # We construct the TEQC args line
        teqcargs = [
            "-O.mo[nument] '{}'".format(self.raw_content["1."]["ID"][:4]),
            "-M.mo[nument] '{}'".format(self.raw_content["1."]["ID"][:4]),
            "-O.mn '{}'".format(self.raw_content["1."]["ID"][:4]),
            "-O.px[WGS84xyz,m] {} {} {}".format(
                self.raw_content["2."]["X coordinate (m)"],
                self.raw_content["2."]["Y coordinate (m)"],
                self.raw_content["2."]["Z coordinate (m)"],
            ),
            "-O.s[ystem] {}".format(o_system),
            "-O.rt '{}'".format(instru["receiver"]["Receiver Type"]),
            "-O.rn '{}'".format(instru["receiver"]["Serial Number"]),
            "-O.rv '{}'".format(instru["receiver"]["Firmware Version"]),
            "-O.at '{}'".format(instru["antenna"]["Antenna Type"]),
            "-O.an '{}'".format(instru["antenna"]["Serial Number"]),
            "-O.pe[hEN,m] {} {} {}".format(
                instru["antenna"]["Marker->ARP Up Ecc. (m)"].zfill(8),
                instru["antenna"]["Marker->ARP East Ecc(m)"].zfill(8),
                instru["antenna"]["Marker->ARP North Ecc(m)"].zfill(8),
            ),
            "-O.o[perator] '{}'".format(
                self.raw_content["11."]["Preferred Abbreviation"]
            ),
            "-O.r[un_by] '{}'".format(
                self.raw_content["11."]["Preferred Abbreviation"]
            ),
            "-O.ag[ency] '{}'".format(
                self.raw_content["12."]["Preferred Abbreviation"]
            ),
        ]

        return teqcargs, ignored

    def find_instru4rnx(self, starttime, endtime, ignore=False, round_instru_dates=False):
        """
        Returns period's metadata in vars and dicts
        fitted for RinexFile modification methods.

        formerly rinex_metadata_lines
        """

        ### searching for the instrumentation corresponding to the period
        instru, ignored = self.find_instru(starttime, endtime, ignore, round_instru_dates)

        ### if no instrumentation found, we return None
        if not instru:
            logger.warning(f"No instru. found for period {starttime}-{endtime}.")
            return None, ignored

        def key_chk_instru(dic_inp, keys_inp):
            """
            Check if all keys in keys_inp are present in dic_inp.
            If not, raise an error.
            """
            for k in keys_inp:
                if k not in dic_inp.keys():
                    errmsg = f"Missing '{k}' for instru. period {starttime}-{endtime} ({dic_inp})"
                    logger.error(errmsg)
                    raise KeyError(errmsg)

        ### get useful values in misc_meta
        fourchar_id = self.site_id4

        misc_meta = self.misc_meta

        key_chk_instru(
            misc_meta,
            [
                "IERS DOMES Number",
                "X coordinate (m)",
                "Y coordinate (m)",
                "Z coordinate (m)",
                "operator",
                "agency",
            ],
        )

        domes_id = misc_meta["IERS DOMES Number"]

        antenna_pos = {
            "X": misc_meta["X coordinate (m)"],
            "Y": misc_meta["Y coordinate (m)"],
            "Z": misc_meta["Z coordinate (m)"],
        }

        agencies = {
            "operator": misc_meta["operator"],
            "agency": misc_meta["agency"],
        }

        ### get useful values in instru_rec
        instru_rec = instru["receiver"]

        key_chk_instru(
            instru_rec,
            ["Serial Number", "Receiver Type", "Firmware Version", "Satellite System"],
        )

        receiver = {
            "serial": instru_rec["Serial Number"],
            "type": instru_rec["Receiver Type"],
            "firmware": instru_rec["Firmware Version"],
        }
        observable_type = instru_rec["Satellite System"]

        ### get useful values in instru_ant
        instru_ant = instru["antenna"]

        key_chk_instru(
            instru_ant,
            [
                "Serial Number",
                "Antenna Type",
                "Marker->ARP Up Ecc. (m)",
                "Marker->ARP East Ecc(m)",
                "Marker->ARP North Ecc(m)",
            ],
        )

        antenna = {
            "serial": instru_ant["Serial Number"],
            "type": instru_ant["Antenna Type"],
        }

        antenna_delta = {
            "H": instru_ant["Marker->ARP Up Ecc. (m)"],
            "E": instru_ant["Marker->ARP East Ecc(m)"],
            "N": instru_ant["Marker->ARP North Ecc(m)"],
        }

        ### final output tuple
        metadata_vars = (
            fourchar_id,
            domes_id,
            observable_type,
            agencies,
            receiver,
            antenna,
            antenna_pos,
            antenna_delta,
        )

        return metadata_vars, ignored

    def rinex_full_history_lines(self):
        """
        Get the sting lines to have the full site history in the RINEX header
        """
        rec_stk = []
        ant_stk = []

        for ins in self.instrus:
            rec_stk.append(ins["receiver"])
            ant_stk.append(ins["antenna"])

        def _stack_lines(instru_stk, instru_name="Receiver"):
            lines_instru_stk = []
            lastl1, lastl2, lastl3 = None, None, None

            inam = instru_name

            for _, inss in enumerate(instru_stk):
                l1 = " ".join((inam, inss[inam + " Type"], str(inss["Serial Number"])))
                l2 = "Installed on " + str(inss["Date Installed"])
                l3 = "Removed on " + str(inss["Date Removed"])

                if l1 == lastl1 and l2 == lastl2 and l3 == lastl3:
                    continue
                else:
                    lines_instru_stk.append(l1)
                    lines_instru_stk.append(l2)
                    lines_instru_stk.append(l3)
                    lastl1 = l1
                    lastl2 = l2
                    lastl3 = l3

            return lines_instru_stk

        return _stack_lines(rec_stk, "Receiver") + _stack_lines(ant_stk, "Antenna")

    def write_json(self, output=None):
        """
        Writes sitelog's dict to json. If no output provided, will write it in
        the same directory as the sitelog.
        """

        filename = os.path.splitext(os.path.basename(self.path))[0]
        if not output:
            output = os.path.dirname(self.path)
        elif not os.path.isdir(output):
            logger.error("error, output folder incorrect " + output)
            return None

        outputfilejson = os.path.join(output, filename + ".json")
        with open(outputfilejson, "w+") as j:
            json.dump(self.raw_content, j, default=str)

        return outputfilejson


def equal_instru(instru1, instru2, compare_dates=False):
    """
    Compare two instrumentation dict. and return True if they are equal
    """

    # Compare dates
    if compare_dates and (instru1["dates"] != instru2["dates"]):
        return False

    # Lambda function to filter out 'Date Installed' and 'Date Removed' keys
    filter_keys = lambda d: {
        k: v for k, v in d.items() if k not in ["Date Installed", "Date Removed"]
    }

    # Compare receiver
    if filter_keys(instru1["receiver"]) != filter_keys(instru2["receiver"]):
        return False

    # Compare antenna
    if filter_keys(instru1["antenna"]) != filter_keys(instru2["antenna"]):
        return False

    return True

    # def stationinfo(self, output = None):
    #     '''
    #     Attempt of writing a function that will write a stationinfo file from a
    #     list of sitelogs. This function writes a list of almost- stationinfo compatible
    #     entries, plus a header. To work, you have to write the XXX-marked lines that are
    #     not yet complient with stationinfo format, and make an external function to
    #     combine the lines from differetn sitelogs and add the header line.
    #     '''
    #
    #     if output:
    #         if not os.path.isdir(output):
    #             print("error, output folder incorrect " + output)
    #             return None
    #
    #     header = []
    #
    #     now = datetime.strftime(datetime.now(), '%Y-%m-%d %H%M')
    #
    #     header.append("# Station.raw_content written by Rinexmod user             on ".format(now))
    #     header.append("* Reference file : station.raw_content")
    #     header.append("*")
    #     header.append("*")
    #     header.append("*SITE  Station Name      Session Start      Session Stop       Ant Ht   HtCod  Ant N    Ant E    Receiver Type         Vers                  SwVer  Receiver SN           Antenna Type     Dome   Antenna SN")
    #
    #     stationinfo = []
    #
    #     SITE = self.raw_content['1.']['Four Character ID']
    #     Station_Name = self.raw_content['1.']['Site Name']
    #
    #     pattern = re.compile(r'[ 0](0+)[0-9]')
    #
    #     for installation in self.instrus:
    #
    #         Session_Start = installation['dates'][0]
    #         # print(type(Session_Start))
    #         Session_Start = datetime.strftime(Session_Start, '%Y %j %H %M %S')
    #         Session_Start = re.sub(pattern, ' ', Session_Start)
    #
    #         # Session_Start = Session_Start.replace(' 00', '   ')
    #         # Session_Start = Session_Start.replace(' 0', '  ')
    #         # Session_Start = Session_Start.replace('  00', '   0')
    #         # Session_Start = Session_Start.replace(' 00', '  0')
    #
    #         Session_Stop = installation['dates'][1]
    #         # print(type(Session_Stop))
    #
    #         # Session_Stop = Session_Stop - timedelta(seconds=1)
    #         Session_Stop = datetime.strftime(Session_Stop, '%Y %j %H %M %S')
    #         Session_Stop = re.sub(pattern, ' ', Session_Stop)
    #         # Session_Stop = Session_Stop.replace(' 0', '  ')
    #         #
    #         # Session_Start = Session_Start.replace('  00', '   0')
    #         # Session_Stop = Session_Stop.replace(' 00', '  0')
    #
    #         Ant_Ht = installation['antenna']['Marker->ARP Up Ecc. (m)']
    #         HtCod = None #installation['antenna'][''] #
    #         Ant_N = installation['antenna']['Marker->ARP North Ecc(m)']
    #         Ant_E = installation['antenna']['Marker->ARP East Ecc(m)']
    #         Receiver_Type = installation['receiver']['Receiver Type']
    #         Vers = None #installation['receiver'][''] #
    #         SwVer = installation['receiver']['Firmware Version']
    #         Receiver_SN = installation['receiver']['Serial Number']
    #         Antenna_Type = installation['antenna']['Antenna Type']
    #         Dome = installation['antenna']['Antenna Radome Type']
    #         Antenna_SN = installation['antenna']['Serial Number']
    #
    #         info_line = [SITE,
    #                      Station_Name,
    #                      Session_Start,
    #                      Session_Stop,
    #                      Ant_Ht,
    #                      HtCod,
    #                      Ant_N,
    #                      Ant_E,
    #                      Receiver_Type,
    #                      Vers,
    #                      SwVer,
    #                      Receiver_SN,
    #                      Antenna_Type,
    #                      Dome,
    #                      Antenna_SN]
    #
    #         stationinfo.append(info_line)
    #
    #     return stationinfo



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
API functions of RinexMod

Created on Wed Mar  8 12:14:54 2023

@author: psakic
"""

import argparse
import os
import re
import subprocess
import string
import pandas as pd
from datetime import datetime, timedelta


import rinexmod as rimo
import rinexmod.common.gamit_meta as rimo_gmm
import rinexmod.classes.metadata as rimo_mda
import rinexmod.classes.rinexfile as rimo_rnx


import rinexmod.logger as rimo_log
logger = rimo_log.logger_define("INFO")

# *****************************************************************************
# define Python user-defined exceptions
class RinexModError(Exception):
    pass


class RinexModInputArgsError(RinexModError):
    pass


class RinexFileError(RinexModError):
    pass


class MetaDataError(RinexModError):
    pass


class ReturnListError(RinexModError):
    pass



# *****************************************************************************
# core functions


def listfiles(directory, extension, recursive=True):
    # returns list of paths
    liste = []
    extension = extension.lower()
    for dirpath, dirnames, files in os.walk(directory):
        for name in files:
            if name.lower().endswith(extension):
                file = os.path.join(dirpath, name)
                liste.append(file)
        if not recursive:
            break
    return list(sorted(liste))


# get Git hash (to get a version number-equivalent of the RinexMod used)
def get_git_hash():
    """
    Gives the Git hash to have a tracking of the used version

    Returns
    -------
    7 characters Git hash
    """
    script_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

    cmd = ["git", "--git-dir", script_path + "/.git", "rev-parse", "--short", "HEAD"]
    try:
        githash = (
            subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
            .decode("ascii")
            .strip()[:7]
        )
    except Exception:
        # logger.warn("unable to get the git commit version")
        githash = "xxxxxxx"

    ####NB: 2msec to run this fuction
    return githash


def make_site_id9(site_id_inp):
    """
    Converts a site ID to a 9-character format.

    This function takes a site ID and converts it to a 9-character format.
    If the input site ID is already 9 characters long, it returns the uppercase version of the input.
    If the input site ID is 4 characters long, it appends '00XXX' to the uppercase version of the input.
    Otherwise, it takes the first 4 characters of the input, converts them to uppercase, and appends '00XXX'.

    Parameters
    ----------
    site_id_inp : str
        The input site ID to be converted.

    Returns
    -------
    str
        The site ID in 9-character format.
    """
    if len(site_id_inp) == 9:
        return site_id_inp.upper()
    elif len(site_id_inp) == 4:
        return site_id_inp.upper() + "00XXX"
    else:
        return site_id_inp[:4].upper() + "00XXX"


# *****************************************************************************
# Metadata import


def metadata_input_manage(sitelog_inp, force=False):
    """
    Manage the multiple types possible for metadata inputs
    Return a list of MetaData to be handeled by find_mda4site

    Possible inputs are:
     * list of string (sitelog file paths),
     * single string (single sitelog file path)
     * single string (directory containing the sitelogs)
     * list of MetaData objects
     * single MetaData object

    Parameters
    ----------
    sitelog_inp : list of str, str, list of MetaData. single MetaData
        Various Input sitelogs.
    force : bool
        Force a single sitelog file path.

    Returns
    -------
        list of MetaData objects.
        (can be a singleton)

    """
    # single MetaData object
    if isinstance(sitelog_inp, rimo_mda.MetaData):
        return [sitelog_inp]
    # list of MetaData objects
    elif isinstance(sitelog_inp, list) and isinstance(
        sitelog_inp[0], rimo_mda.MetaData
    ):
        return sitelog_inp
    # single string or list of string
    elif isinstance(sitelog_inp, str) or isinstance(sitelog_inp, list):
        return sitlgs2mda_objs(
            sitelog_inp, force=force, return_list_even_if_single_input=True
        )
    else:
        logger.error(
            "Wrong Input, must be a list of string (path), a single string (path),"
            "a MetaData object, or a list of MetaData objects. Input given: %s, %s",
            sitelog_inp,
            type(sitelog_inp),
        )
        raise RinexModInputArgsError


def gamit2mda_objs(
    station_info_inp,
    lfile_inp=None,
    force_fake_coords=False,
    ninecharfile_inp=None,
    rev=False,
):
    """
    Read a GAMIT files and convert their content to MetaData objects

    Parameters
    ----------
    station_info_inp : str or pd.DataFrame
        Path of a GAMIT station.info file to obtain
        GNSS site metadata information.
    lfile_inp : str or pd.DataFrame
        Path of a GAMIT apriori apr/L-File to obtain
        GNSS site position and DOMES information.
    force_fake_coords : bool, optional
        hen using GAMIT station.info metadata without apriori coordinates in
        the L-File, gives fake coordinates at (0°,0°) to the site.
        The default is False.
    ninecharfile_inp : str, optional
        Path of a file that contains 9-char. site names.
        The default is None.
    rev : bool, optional
        reverse order of inputs for station.info filled from the
        newest to the oldest change
        (Automatic sort of the station.info file as DataFrame
         is then disabled)
        The default is False.

    Returns
    -------
    mdaobjs_lis : list
        list of MetaData objects.

    """
    if isinstance(station_info_inp, pd.DataFrame):
        df_stinfo_raw = station_info_inp
        stinfo_name = "station.info"
    else:
        df_stinfo_raw = rimo_gmm.read_gamit_station_info(
            station_info_inp, sort=False
        )  # sort = not rev if rev, no sort
        stinfo_name = os.path.basename(station_info_inp)

    if not lfile_inp:
        df_apr = pd.DataFrame(columns=["site"])
        logger.warning(
            "No L-File provided, fake coordinates will be used! "
            "(force_fake_coords forced to True)"
        )
        force_fake_coords = True
    elif isinstance(lfile_inp, pd.DataFrame):
        df_apr = lfile_inp
    else:
        df_apr = rimo_gmm.read_gamit_apr_lfile(lfile_inp)

    if not ninecharfile_inp is None:
        nine_char_dict = rimo.rinexmod_api.read_ninecharfile(ninecharfile_inp)
    else:
        nine_char_dict = dict()

    sites_isin = df_stinfo_raw["site"].isin(df_apr["site"])
    ### for the stats only
    sites_uniq = pd.Series(df_stinfo_raw["site"].unique())
    sites_isin_uniq = sites_uniq.isin(df_apr["site"].unique())
    n_sites_notin = len(sites_uniq) - sum(sites_isin_uniq)

    if n_sites_notin > 0 and not force_fake_coords:
        logger.warning(
            "%i/%i sites in %s are not in apr/lfile. They are skipped (you can force fake coords with -fc)",
            n_sites_notin,
            len(sites_uniq),
            stinfo_name,
        )
        df_stinfo = df_stinfo_raw[sites_isin]
    elif n_sites_notin > 0 and force_fake_coords:
        logger.warning(
            "%i/%i sites in %s are not in apr/lfile. Fake coords at (0°,0°) used",
            n_sites_notin,
            len(sites_uniq),
            stinfo_name,
        )
        df_stinfo = df_stinfo_raw
    else:  #### no missing coords, n_sites_notin == 0
        df_stinfo = df_stinfo_raw

    df_stinfo_grp = df_stinfo.groupby("site")

    mdaobjs_lis = []

    logger.info("%i sites will be extracted from %s", len(df_stinfo_grp), stinfo_name)

    for site, site_info in df_stinfo_grp:
        logger.debug("extract %s from %s", site, stinfo_name)

        #### search in ninecharfile IMPROVE ME !!!
        if site in nine_char_dict.keys():
            site_use = nine_char_dict[site]
            logger.debug("4 > 9 char. conversion: %s > %s", site, site_use)
        else:
            site_use = site

        mdaobj = rimo_mda.MetaData(sitelogfile=None)
        mdaobj.set_from_gamit(
            site_use,
            df_stinfo,
            df_apr,
            force_fake_coords=force_fake_coords,
            station_info_name=stinfo_name,
        )
        if rev:
            mdaobj.instrus.reverse()
            logger.info("Reversing order of instrumental changes")
        mdaobjs_lis.append(mdaobj)

    logger.info("%i sites have been extracted from %s", len(mdaobjs_lis), stinfo_name)

    return mdaobjs_lis


def sitlgs2mda_objs(
    sitelog_filepath, force=False, return_list_even_if_single_input=True
):
    """
    Read a set of sitelog files and convert them to MetaData objects

    Parameters
    ----------
    sitelog_filepath : str or list of str
        path of a single sitelog file or a set of sitelogs (stored in a list).
    force : bool, optional
        if it is not possible to read a sitelog (bad format), we skip it.
        The default is False.
    return_list_even_if_single_input : bool, optional
        if a single sitelog is given, return the corresponding MetaData object
        into a list (singleton). The default is True.

    Raises
    ------
    MetaDataError
    RinexModInputArgsError

    Returns
    -------
    mdaobjs_lis : list
        list of MetaData objects.

    """

    #### differenciating cases
    # Case of a list of sitelogs
    if type(sitelog_filepath) is list:
        slgs_all = sitelog_filepath
        sitelog_filepath = "input sitelog list"
    # Case of one single sitelog
    elif os.path.isfile(sitelog_filepath):
        slgs_all = [sitelog_filepath]
    # Case of a folder
    elif os.path.isdir(sitelog_filepath):
        sitelog_extension = ".log"
        slgs_all = listfiles(sitelog_filepath, sitelog_extension)

        sitelog_pattern = re.compile(r"\w{4,9}_\d{8}.log")
        slgs_all = [f for f in slgs_all if sitelog_pattern.match(os.path.basename(f))]
    ### case of no file nor folder
    else:
        logger.error(
            "unable to handle file/directory. Does it exists?: %s", sitelog_filepath
        )
        raise RinexModInputArgsError

    #### Read the sitelogs
    logger.info("**** %i sitelogs detected (in %s)", len(slgs_all), sitelog_filepath)
    for sl in slgs_all:
        logger.debug(sl)
    # Get last version of sitelogs if multiple available
    # (1st gross search based on date in filename)
    slgs_latest = _slg_find_latest_name(slgs_all)
    # load the sitelogs as metadata objects
    mdaobjs_lis, slgs_bad_lis = load_sitelogs(slgs_latest, force)
    # Get last version of sitelogs if multiple available
    # (2nd fine search based on date in "Date Prepared" field)
    mdaobjs_lis = _mda_find_latest_prep(mdaobjs_lis)

    logger.info("**** %i most recent sitelogs selected", len(mdaobjs_lis))

    for sl in mdaobjs_lis:
        logger.debug(sl.path)

    if len(slgs_bad_lis) > 0:
        logger.warning("**** %i badly-parsed & ignored sitelogs", len(slgs_bad_lis))

    if len(mdaobjs_lis) <= 1 and not return_list_even_if_single_input:
        mdaobjs_lis = mdaobjs_lis[0]

    return mdaobjs_lis


def rinexs2mda_objs(rinex_paths, ninecharfile_inp=None):
    """
    Read a set of RINEX files and convert them to MetaData objects

    Parameters
    ----------
    rinex_paths : list of str
        path of a single rinex file or a set of rinex files (stored in a list).

    ninecharfile_inp : str, optional
        Path of a file that contains 9-char. site names.
        The default is None.

    Returns
    -------
    mdaobjs_lis : list
        list of MetaData objects.

    """

    if not isinstance(rinex_paths, list):
        rinex_paths = [rinex_paths]

    if not ninecharfile_inp is None:
        nine_char_dict = read_ninecharfile(ninecharfile_inp)
    else:
        nine_char_dict = dict()

    mda_stk = []
    for rnx in sorted(rinex_paths):
        mda = rimo_mda.MetaData()
        mda.set_from_rinex(rnx)

        #### search in ninecharfile IMPROVE ME !!!
        if mda.site_id9[-3:] == "XXX" and mda.site_id4 in nine_char_dict.keys():
            site_id4_orig = mda.site_id4
            mda.site_id = nine_char_dict[mda.site_id4]
            logger.info(f"4 > 9 char. conversion: {site_id4_orig:} > {mda.site_id9:}")

        mda_stk.append(mda)

    mdaobjs_lis, mdaobj_dic = group_mda(mda_stk)

    return mdaobjs_lis


def group_mda(mdaobj_list_inp):
    """
    Group a list of MetaData objects into a dictionary and return a list and dictionary of merged objects.

    This function processes a list of MetaData objects, grouping them by their `site_id9` attribute.
    If multiple MetaData objects share the same `site_id9`, their instrumentation data is merged.

    Parameters
    ----------
    mdaobj_list_inp : list
        A list of MetaData objects to be grouped.

    Returns
    -------
    mdaobj_lis : list
      A list of grouped MetaData objects.
    mdaobj_dic : dict
        A dictionary where keys are `site_id9` and values are the grouped MetaData objects.
    """
    mdaobj_dic = dict()  # Initialize an empty dictionary to store grouped MetaData objects.
    for mda in mdaobj_list_inp:
        # Check if the current MetaData object's site_id9 is already in the dictionary.
        if not mda.site_id9 in mdaobj_dic.keys():
            # If not, add it to the dictionary.
            mdaobj_dic[mda.site_id9] = mda
        else:
            # If it exists, merge the instrumentation data from the current object.
            for inst in mda.instrus:
                mdaobj_dic[mda.site_id9].add_instru(
                    inst["receiver"],
                    inst["antenna"],
                    inst["dates"][0],
                    inst["dates"][1],
                )

    # Convert the dictionary values to a list of grouped MetaData objects.
    mdaobj_lis = [v for k, v in mdaobj_dic.items()]

    return mdaobj_lis, mdaobj_dic


def load_sitelogs(sitelogs_inp, force=False):
    """
    Process a list of sitelogs and return a list of MetaData objects and a list of bad sitelogs.

    Parameters
    ----------
    sitelogs_inp : list
        List of paths to the latest sitelog files.
    force : bool, optional
        If True, force processing even if sitelog is not parsable. Default is False.

    Returns
    -------
    mdaobjs_lis : Iterable
        List of MetaData objects.
    bad_sitelogs_lis : Iterable
        List of sitelogs that could not be parsed.
    """

    def _load_slg(sitelog_filepath, force_load):
        # Creating MetaData object
        try:
            mdaobj_load = rimo_mda.MetaData(sitelog_filepath)
        except Exception as e:
            # If sitelog is not parsable
            logger.error(
                "The sitelog is not parsable: %s (%s)",
                os.path.basename(sitelog_filepath),
                str(e),
            )
            if not force_load:
                raise MetaDataError
            else:
                mdaobj_load = None

        return mdaobj_load

    mdaobjs_lis = []
    bad_sitelogs_lis = []
    for sta_sitelog in sitelogs_inp:
        mdaobj = _load_slg(sta_sitelog, force)
        # Appending to list
        if mdaobj:
            mdaobjs_lis.append(mdaobj)
        else:
            bad_sitelogs_lis.append(sta_sitelog)

    return mdaobjs_lis, bad_sitelogs_lis


def _slg_find_latest_name(all_sitelogs_filepaths):
    """
    Find the latest version of a sitelog within a list of sitelogs,
    based on date in its filename
    (mainly for time consumption reduction)

    see also _mda_find_latest_prep (more reliable but slower)
    """
    # We list the available sites to group sitelogs
    bnm = os.path.basename
    sl_bnm = [bnm(sl) for sl in all_sitelogs_filepaths]
    sl_sta = [sl[0:4] for sl in sl_bnm]

    # set the output latest sitelog list
    latest_sitelogs_filepaths = []

    for sta in sl_sta:
        # Grouping by site
        sta_sitelogs = [
            slp for (slp, sln) in zip(all_sitelogs_filepaths, sl_bnm) if sln[0:4] == sta
        ]
        # Getting dates from basename
        # sitelogs_dates0 = [os.path.splitext(bnm(sl))[0][-8:] for sl in sta_sitelogs]
        # Getting dates from basename and parsing 'em
        date_from_fn = lambda x: os.path.splitext(bnm(x))[0][-8:]
        sitelogs_dates = []
        for sl in sta_sitelogs:
            try:
                d = datetime.strptime(date_from_fn(sl), "%Y%m%d")
                sitelogs_dates.append(d)
            except ValueError as e:
                logger.error(
                    f"bad date {date_from_fn(sl):} in sitelog's filename: {sl:}"
                )
                raise e
        # We get the max date and put it back to string format.
        maxdate = max(sitelogs_dates).strftime("%Y%m%d")
        # We filter the list with the max date string, and get a one entry list, then transform it to string
        sta_sitelog = [sl for sl in sta_sitelogs if maxdate in date_from_fn(sl)][0]

        latest_sitelogs_filepaths.append(sta_sitelog)

    return latest_sitelogs_filepaths


def _mda_find_latest_prep(mdaobjs_inp):
    """
    Find the latest version of a MetaData object within a list of MetaData objects,
    based on preparation date

    more reliable than _slg_find_latest_name but requires that the metadata objects have been loaded
    """
    # We list the available sites to group sitelogs
    mdaobjs_lis = mdaobjs_inp
    # mdaobjs_lis = sorted(mdaobjs_lis, key=lambda x: x.misc_meta["date prepared"], reverse=True)

    # set the output latest sitelog list
    mdaobjs_latest = []

    sites_all = list(sorted(list(set([md.site_id4 for md in mdaobjs_lis]))))

    for site in sites_all:
        # Grouping by site
        mdaobjs_site = [m for m in mdaobjs_lis if m.site_id4 == site]
        # Getting dates from basename and parsing 'em
        mdaobjs_site_dates = [m.misc_meta["date prepared"] for m in mdaobjs_site]
        # We get the max date and put it back to string format.
        maxdate = max(mdaobjs_site_dates)
        # We filter the list with the max date string, and get a one entry list, then transform it to string
        mdaobj_latest = [
            md for md in mdaobjs_site if md.misc_meta["date prepared"] == maxdate
        ][0]

        mdaobjs_latest.append(mdaobj_latest)

    return mdaobjs_latest


def find_mda4site(rnxobj_or_site4char, mdaobjs_lis, force):
    """
    Finding the right MetaData object

    If is list, can not use force.
    If no metadata found, do not process.
    """
    if type(rnxobj_or_site4char) is str:
        rnx_4char = rnxobj_or_site4char[:4]
        err_label = rnxobj_or_site4char
    else:
        rnx_4char = rnxobj_or_site4char.get_site(True, True)
        err_label = rnxobj_or_site4char.filename

    logger.debug("Searching corresponding metadata for site: " + rnx_4char)

    if len(mdaobjs_lis) == 0:
        logger.warning("The metadata list provided is empty!")

    mdaobj = None
    if rnx_4char not in [mda.site_id4 for mda in mdaobjs_lis]:
        if len(mdaobjs_lis) == 1:
            if not force:
                errmsg = "33 - RINEX name's site does not correspond to provided metadata, use -f option to force"
                logger.error(f"{errmsg:110s} - {err_label:}")
                raise RinexModInputArgsError
            else:
                errmsg = "34 - RINEX name's site does not correspond to provided metadata, forced processing anyway"
                logger.warning(f"{errmsg:110s} - {err_label:}")
        else:
            errmsg = "33 - No metadata found for this RINEX"
            logger.error(f"{errmsg:110s} - {err_label:}")
            raise RinexModInputArgsError
    else:
        mdaobjs_site = [md for md in mdaobjs_lis if md.site_id4 == rnx_4char]
        if len(mdaobjs_site) == 1:
            mdaobj = mdaobjs_site[0]
        else:
            # the assumption that latest sitelog has been found in sitlgs2mda_objs is wrong!
            # a second search with the preparation date is performed here
            mdaobj = _mda_find_latest_prep(mdaobjs_site)[0]

    return mdaobj


def apply_mda2rnxobj(rnxobj, mdaobj, ignore=False, keep_rnx_rec=False, round_instru_dates=False):
    """
    apply a MetaData object on a RinexFile object
    to modify this RinexFile with the rights metadata
    """

    ### do this check with 9 chars at one point
    rnx_4char = rnxobj.get_site(True, True)
    # Site name from the sitelog
    mda_4char = mdaobj.site_id4 # misc_meta["ID"].lower()[:4]

    if rnx_4char != mda_4char:
        logger.warning(
            "RINEX and metadata 4 char. codes do not correspond, "
            "but I assume you know what you are doing (%s,%s)",
            rnx_4char,
            mda_4char,
        )

    # Get rinex header values from sitelog infos and start and end time of the file
    # ignore option is to ignore firmware changes between instrumentation periods.
    mda_vars, ignored = mdaobj.find_instru4rnx(
        rnxobj.start_date, rnxobj.end_date, ignore, round_instru_dates
    )

    if not mda_vars:
        logger.error(
            "35 - No instrumentation corresponding to the RINEX epoch - %s",
            rnxobj.filename,
        )
        raise MetaDataError

    if ignored:
        logger.warning(
            "36 - Instrumentation comes from merged metadata periods with different firmwares, processing anyway - %s",
            rnxobj.filename,
        )

    (
        fourchar_id,
        domes_id,
        sat_sys_long_fmt,
        agencies,
        rec,
        antenna,
        ant_pos,
        ant_delta,
    ) = mda_vars

    ## Apply the modifications to the RinexFile object
    rnxobj.mod_marker(fourchar_id, domes_id)
    rnxobj.mod_receiver(keep_rnx_rec=keep_rnx_rec, **rec)
    rnxobj.mod_interval(rnxobj.sample_rate_numeric)
    rnxobj.mod_antenna(**antenna)
    rnxobj.mod_antenna_pos(**ant_pos)
    rnxobj.mod_antenna_delta(**ant_delta)
    rnxobj.mod_agencies(**agencies)
    if not keep_rnx_rec:
        ### if keep_rnx_rec is active, we keep the sat system in the header
        rnxobj.mod_sat_system(sat_sys_long_fmt)

    return rnxobj


# *****************************************************************************
# modification keyword dictionary functions


def modif_kw_check(modif_kw):
    """
    Check if acceptable modification keywords have been provided

    Raise a RinexModInputArgsError Exception if not
    """

    acceptable_keywords = [
        "station",
        "marker_name",
        "marker_number",
        "receiver_serial",
        "receiver_type",
        "receiver_fw",
        "antenna_serial",
        "antenna_type",
        "antenna_X_pos",
        "antenna_Y_pos",
        "antenna_Z_pos",
        "antenna_H_delta",
        "antenna_E_delta",
        "antenna_N_delta",
        "operator",
        "agency",
        "sat_system",
        "observables",
        "interval",
        "filename_data_freq",
        "filename_file_period",
        "filename_data_source",
        "comment(_[0-9]+)?",
    ]
    ### comment is a regex, bc several comments are possible
    # suffix _N is added by ParseKwargs
    # but comment without suffix must remain possible (for API mode)

    for kw in modif_kw:
        if not any([re.match(akw, kw) for akw in acceptable_keywords]):
            logger.error(
                "'{}' is not an acceptable keyword for header modification.".format(kw)
            )
            return RinexModInputArgsError

    return None


def apply_modifkw2rnxobj(rnxobj, modif_kw):
    """
    apply a modification keywords on a RinexFile object
    to modify this RinexFile with the rights metadata
    """

    def __keys_in_modif_kw(keys_in):
        return all([e in modif_kw.keys() for e in keys_in])

    rnxobj.mod_marker(modif_kw.get("marker_name"), modif_kw.get("marker_number"))

    # legacy keyword, 'marker_name' should be used instead
    rnxobj.mod_marker(modif_kw.get("station"))

    rnxobj.mod_receiver(
        modif_kw.get("receiver_serial"),
        modif_kw.get("receiver_type"),
        modif_kw.get("receiver_fw"),
    )

    rnxobj.mod_antenna(modif_kw.get("antenna_serial"), modif_kw.get("antenna_type"))

    rnxobj.mod_antenna_pos(
        modif_kw.get("antenna_X_pos"),
        modif_kw.get("antenna_Y_pos"),
        modif_kw.get("antenna_Z_pos"),
    )

    rnxobj.mod_antenna_delta(
        modif_kw.get("antenna_H_delta"),
        modif_kw.get("antenna_E_delta"),
        modif_kw.get("antenna_N_delta"),
    )

    rnxobj.mod_agencies(modif_kw.get("operator"), modif_kw.get("agency"))

    rnxobj.mod_sat_system(modif_kw.get("sat_system"))
    # legacy keyword, 'sat_system' should be used instead
    rnxobj.mod_sat_system(modif_kw.get("observables"))

    rnxobj.mod_interval(modif_kw.get("interval"))

    # for the filename
    rnxobj.mod_filename_file_period(modif_kw.get("filename_file_period"))
    rnxobj.mod_filename_data_freq(modif_kw.get("filename_data_freq"))
    rnxobj.mod_filename_data_source(modif_kw.get("filename_data_source"))

    # comment
    # special case: several keys comment_1, comment_2, comment_N are possible
    # number are added automatically by ParseKwargs
    comment_keys = [k for k in modif_kw.keys() if "comment" in k]
    for ck in comment_keys:
        rnxobj.add_comment(modif_kw.get(ck).split("_")[0])

    return rnxobj


# *****************************************************************************
# dictionnary as output for gnss_delivery workflow

def rtun_lsts_make(rnxobj_or_dict, return_lists=dict()):
    """
    Construct the so called ``return_lists`` (which are actually dictionnaries)

    return_lists have the structure:

    ```
    return_lists[major_rinex_version][sample_rate_string][file_period]
    ```

    the input ``return_lists`` is populated with the content of the input
    ``rnxobj_or_dict``

    Parameters
    ----------
    rnxobj_or_dict : RinexFile object or dict

        The input new information which will populate the return_lists

        if a dict is provided, we assume it is a singleton with a standard
        return_lists structure to be merged in a bigger return_lists

    return_lists : dict, optional
        A potential pre-exisiting return_lists.
        Per default it is a brand-new return_lists for scratch.
        The default is dict().

    Returns
    -------
    return_lists : dict
        The input return_lists populated with the input rnxobj_or_dict.

    Note
    ----
    Specific usage for the IPGP's gnss_delivery workflow

    """

    if isinstance(rnxobj_or_dict, rimo_rnx.RinexFile):
        rnxobj = rnxobj_or_dict
        major_rinex_version = rnxobj.version[0]
        sample_rate_string = rnxobj.sample_rate_string
        file_period = rnxobj.file_period
        path_output = rnxobj.path_output
    elif isinstance(rnxobj_or_dict, dict):
        rtrnlst = rnxobj_or_dict
        major_rinex_version = list(rtrnlst.keys())[0]
        sample_rate_string = list(rtrnlst[major_rinex_version].keys())[0]
        file_period = list(rtrnlst[major_rinex_version][sample_rate_string].keys())[0]
        path_output = rtrnlst[major_rinex_version][sample_rate_string][file_period][0]
    else:
        logger.error(
            "Wrong Input, must be RinexFile object or dict. Input given: %s, %s",
            rnxobj_or_dict,
            type(rnxobj_or_dict),
        )
        raise ReturnListError

    # Dict ordered as : RINEX_VERSION, SAMPLE_RATE, FILE_PERIOD
    if major_rinex_version not in return_lists:
        return_lists[major_rinex_version] = {}
    if sample_rate_string not in return_lists[major_rinex_version]:
        return_lists[major_rinex_version][sample_rate_string] = {}
    if file_period not in return_lists[major_rinex_version][sample_rate_string]:
        return_lists[major_rinex_version][sample_rate_string][file_period] = []

    return_lists[major_rinex_version][sample_rate_string][file_period].append(
        path_output
    )

    return return_lists


def rtun_lsts_write(return_lists, logfolder, now_dt=None):
    # Writing an output file for each RINEX_VERSION, SAMPLE_RATE, FILE_PERIOD lists
    if not now_dt:
        now_dt = datetime.now()

    this_outputfile = ""

    for rinex_version in return_lists:
        for sample_rate in return_lists[rinex_version]:
            for file_period in return_lists[rinex_version][sample_rate]:
                this_outputfile = "_".join(
                    [
                        "RINEX" + rinex_version,
                        sample_rate,
                        file_period,
                        datetime.strftime(now_dt, "%Y%m%d%H%M"),
                        "delivery.lst",
                    ]
                )
                this_outputfile = os.path.join(logfolder, this_outputfile)

                # Writting output to temporary file and copying it them to target files
                with open(this_outputfile, "w") as f:
                    f.writelines(
                        "{}\n".format(line)
                        for line in return_lists[rinex_version][sample_rate][
                            file_period
                        ]
                    )
                    logger.debug("Output rinex list written to " + this_outputfile)

    return this_outputfile


def read_ninecharfile(ninecharfile_inp):
    nine_char_dict = dict()

    if isinstance(ninecharfile_inp, str):
        with open(ninecharfile_inp, "r") as F:
            nine_char_list = F.readlines()
    elif isinstance(ninecharfile_inp, list):
        nine_char_list = ninecharfile_inp
    else:
        nine_char_list = list(ninecharfile_inp)

    for site_key in nine_char_list:
        nine_char_dict[site_key[:4].lower()] = site_key.strip()

    return nine_char_dict

# *****************************************************************************
# low level functions originaly from RinexFile class
def search_idx_val(data, field):
    """
    find the index (line number) of a researched field in the RINEX data
    return None if nothing has beeen found
    """
    idx = -1
    out_idx = None
    for e in data:
        idx += 1
        if field in e:
            out_idx = idx
            break
    return out_idx


def slice_list(seq, num):
    """make sublist of num elts of a list"""
    # http://stackoverflow.com/questions/4501636/creating-sublists
    return [seq[i : i + num] for i in range(0, len(seq), num)]


def round_time(dt=None, date_delta=timedelta(minutes=60), to="average"):
    """
    Round a datetime object to a multiple of a timedelta
    dt : datetime.datetime object, default now.
    dateDelta : timedelta object, we round to a multiple of this, default 1 minute.
    to : up / down / average
    from:  http://stackoverflow.com/questions/3463930/how-to-round-the-minute-of-a-datetime-object-python
    """
    round_to = date_delta.total_seconds()
    if dt is None:
        dt = datetime.now()
    seconds = (dt - dt.min).seconds

    if seconds % round_to == 0 and dt.microsecond == 0:
        rounding = (seconds + round_to / 2) // round_to * round_to
    else:
        if to == "up":
            # // is a floor division, not a comment on following line (like in javascript):
            rounding = (
                (seconds + dt.microsecond / 1000000 + round_to) // round_to * round_to
            )
        elif to == "down":
            rounding = seconds // round_to * round_to
        else:
            rounding = (seconds + round_to / 2) // round_to * round_to

    return dt + timedelta(0, rounding - seconds, -dt.microsecond)


def regex_pattern_rinex_filename():
    """
    return a dictionnary with the different REGEX patterns to describe a RIENX filename
    """
    pattern_dic = dict()
    # pattern_dic["shortname"] = "....[0-9]{3}(\d|\D)\.[0-9]{2}(o|d)(|\.(Z|gz))"
    pattern_dic["shortname"] = (
        r"....[0-9]{3}(\d|\D)([0-9]{2}\.|\.)[0-9]{2}(o|d)(|\.(Z|gz))"  ### add subhour starting min
    )
    pattern_dic["longname"] = (
        r".{4}[0-9]{2}.{3}_(R|S|U)_[0-9]{11}_([0-9]{2}\w)_[0-9]{2}\w_\w{2}\.\w{3}(\.gz|)"
    )
    pattern_dic["longname_gfz"] = (
        r".{4}[0-9]{2}.{3}_[0-9]{8}_.{3}_.{3}_.{2}_[0-9]{8}_[0-9]{6}_[0-9]{2}\w_[0-9]{2}\w_[A-Z]*\.\w{3}(\.gz)?"
    )

    return pattern_dic


def dates_from_rinex_filename(rnx_inp):
    """
    determine the start epoch, the end epoch and the period of a RINEX
    file based on its name only.
    The RINEX is not readed. This function is much faster but less reliable
    than the RinexFile.start_date and RinexFile.end_date attribute

    return the start epoch and end epoch as datetime
    and the period as timedelta
    """
    pattern_dic = regex_pattern_rinex_filename()

    pattern_shortname = re.compile(pattern_dic["shortname"])
    pattern_longname = re.compile(pattern_dic["longname"])
    pattern_longname_gfz = re.compile(pattern_dic["longname_gfz"])

    rinexname = os.path.basename(rnx_inp)

    def _period_to_timedelta(peri_inp):
        peri_val = int(peri_inp[0:2])
        peri_unit = str(peri_inp[2])

        if peri_unit == "M":
            unit_sec = 60
        elif peri_unit == "H":
            unit_sec = 3600
        elif peri_unit == "D":
            unit_sec = 86400
        else:
            logger.warn("odd RINEX period: %s, assume it as 01D", peri_inp)
            unit_sec = 86400

        return timedelta(seconds=peri_val * unit_sec)

    ##### LONG rinex name
    if re.search(pattern_longname, rinexname):
        date_str = rinexname.split("_")[2]
        period_str = rinexname.split("_")[3]

        yyyy = int(date_str[:4])
        doy = int(date_str[4:7])
        hh = int(date_str[7:9])
        mm = int(date_str[9:11])
        dt_srt = datetime(yyyy, 1, 1) + timedelta(
            days=doy - 1, seconds=hh * 3600 + mm * 60
        )
        period = _period_to_timedelta(period_str)
        dt_end = dt_srt + period

        return dt_srt, dt_end, period

    ##### LONG rinex name -- GFZ's GODC internal name
    elif re.search(pattern_longname_gfz, rinexname):
        date_str = rinexname.split("_")[5]
        time_str = rinexname.split("_")[6]
        period_str = rinexname.split("_")[7]

        yyyy = int(date_str[:4])
        mo = int(date_str[4:6])
        dd = int(date_str[6:8])

        hh = int(time_str[0:2])
        mm = int(time_str[2:4])
        ss = int(time_str[4:6])

        dt_srt = datetime(yyyy, mo, dd, hh, mm, ss)
        period = _period_to_timedelta(period_str)
        dt_end = dt_srt + period

        return dt_srt, dt_end, period

    ##### SHORT rinex name
    elif re.search(pattern_shortname, rinexname):
        alphabet = list(string.ascii_lowercase)

        doy = int(rinexname[4:7])
        yy = int(rinexname[9:11])

        if yy > 80:
            year = yy + 1900
        else:
            year = yy + 2000

        if rinexname[7] in alphabet:
            h = alphabet.index(rinexname[7])
            period = timedelta(seconds=3600)
        else:
            h = 0
            period = timedelta(seconds=86400)

        dt_srt = datetime(year, 1, 1) + timedelta(days=doy - 1, seconds=h * 3600)
        dt_end = dt_srt + period
        return dt_srt, dt_end, period

    else:
        logger.error("%s has not a RINEX name well formatted", rinexname)
        return None, None, None


def file_period_from_timedelta(start_date, end_date):
    """
    return the RINEX file period (01H, 01D, 15M...) based on a
    start and end date

    Parameters
    ----------

    Returns
    -------
    file_period : str
        file period (01H, 01D, 15M...)
    session : bool
        True if the timedelta refers to a session (<01D)
        False otherwise (01D).

    """
    rndtup = lambda x, t: round_time(x, timedelta(minutes=t), "up")
    rndtdown = lambda x, t: round_time(x, timedelta(minutes=t), "down")
    rndtaver = lambda x, t: round_time(x, timedelta(minutes=t), "average")
    # rounded at the hour
    # maximum and average delta between start and end date
    delta_max = rndtup(end_date, 60) - rndtdown(start_date, 60)
    delta_ave = rndtaver(end_date, 60) - rndtaver(start_date, 60)

    hours_ave = int(delta_ave.total_seconds() / 3600)
    delta_sec = (end_date - start_date).total_seconds()

    # first, the special case : N *full* hours
    if delta_max <= timedelta(seconds=86400 - 3600) and hours_ave > 0:  ## = 23h max
        # delta_ave is a more precise delta than delta_max (average)
        file_period = str(hours_ave).zfill(2) + "H"
        session = True
    # more regular cases : 01H, 01D, nnM, or Unknown
    elif delta_max <= timedelta(seconds=3600):
        # Here we consider sub hourly cases
        session = True
        file_period = None
        for m in [5, 10, 15, 20, 30]:
            if (m * 60 - 1) <= delta_sec <= (m * 60 + 1):
                file_period = str(m).zfill(2) + "M"
        if not file_period:
            # NB: this test is useless, it is treated by the previous test
            file_period = "01H"
    elif hours_ave == 0 and delta_max > timedelta(seconds=3600):  # Note 2
        hours_max = int(delta_max.total_seconds() / 3600)
        file_period = str(hours_max).zfill(2) + "H"
        session = True
    elif (
        timedelta(seconds=3600) < delta_max <= timedelta(seconds=86400 + 3600)
    ):  # Note1
        file_period = "01D"
        session = False
    else:
        file_period = "00U"
        session = False
    # Note1: a tolerance of +/- 1 hours is given because old ashtech RINEXs
    #        includes the epoch of the next hour/day
    #        and then the present delta_max value reach 25
    #        it justifies also the necessity of the delta_ave variable

    # Note 2: very rare (but possible) case --' :
    #         very short file, riding in between* two hours ("a cheval sur 2 heures")
    #         met e.g. for "2024-08-13 11:54:00" > "2024-08-13 12:04:00"
    #         then the delta_max is 2H and the delta_ave is 0
    #         and we must introduce hours_max rather than hours_ave

    return file_period, session



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 27/06/2025 15:10:30

@author: psakic
"""


import warnings
warnings.warn(
    f"from rinexmod import metadata is deprecated, use import rinexmod.classes.metadata instead.",
    category=DeprecationWarning,
    stacklevel=2,
)


from .classes.metadata import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 26/06/2025 18:24:11

@author: psakic
"""

import os
from datetime import datetime
import hatanaka

import rinexmod as rimo
import rinexmod.common.gamit_meta as rimo_gmm
import rinexmod.classes.metadata as rimo_mda
import rinexmod.classes.rinexfile as rimo_rnx
import rinexmod.api.core_fcts as rimo_cor


import rinexmod.logger as rimo_log
logger = rimo_log.logger_define("INFO")


# *****************************************************************************
# Main function


def rinexmod(
    rinexfile,
    outputfolder,
    sitelog=None,
    modif_kw=dict(),
    marker="",
    country="",
    longname=False,
    shortname=False,
    force_rnx_load=False,
    force_sitelog=False,
    ignore=False,
    ninecharfile=None,
    no_hatanaka=False,
    compression="gz",
    relative="",
    verbose=True,
    full_history=False,
    filename_style="basic",
    return_lists=None,
    station_info=None,
    lfile_apriori=None,
    force_fake_coords=False,
    remove=False,
    keep_rnx_rec=False,
    round_instru_dates=False,
):
    """
    Parameters
    ----------
    rinexfile : str
        Input RINEX file to process.
    outputfolder : str
        Folder where to write the modified RINEX files.
    sitelog : str, list of str, MetaData object, list of MetaData objects, optional
        Get the RINEX header values from a sitelog.
        Possible inputs are:
         * list of string (sitelog file paths),
         * single string (single sitelog file path or directory containing the sitelogs),
         * list of MetaData object
         * single MetaData object
        The function will search for the latest and right sitelog
        corresponding to the site.
        One can force a single sitelog with force_sitelog.
        The default is None.
    modif_kw : dict, optional
        Modification keywords for RINEX's header fields and/or filename.
        Will override the information from the sitelog.
        Acceptable keywords for the header fields:
         * comment
         * marker_name
         * marker_number
         * station (legacy alias for marker_name)
         * receiver_serial
         * receiver_type
         * receiver_fw
         * antenna_serial
         * antenna_type
         * antenna_X_pos
         * antenna_Y_pos
         * antenna_Z_pos
         * antenna_H_delta
         * antenna_E_delta
         * antenna_N_delta
         * operator
         * agency
         * sat_system (M, G, R, E, C...)
         * observables (legacy alias for sat_system)
         * interval
        Acceptable keywords for the filename:
         * filename_file_period (01H, 01D...)
         * filename_data_freq (30S, 01S...)
         * filename_data_source (R, S, U)
        The default is dict().
    marker : str, optional
        A four or nine character site code that will be used to rename
        input files.
        Apply also to the header's MARKER NAME,
        but a custom modification keyword marker_name='XXXX' overrides it
        (modif_kw argument below)
        The default is ''.
    country : str, optional
        A three character string corresponding to the ISO 3166 Country code
        that will be used to rename input files.
        It overrides other country code sources (sitelog, --marker...)
        list of ISO country codes:
        https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes
        The default is ''.
    longname : bool, optional
        Force RINEX file renaming with long name convention (force gzip compression).
        Mutually exclusive with shortname.
        The default is False.
    shortname : bool, optional
        Force RINEX file renaming with short name convention.
        Mutually exclusive with longname.
        The default is False.
    force_rnx_load : bool, optional
        Force the loading of the input RINEX. Useful if its name is not standard.
        The default is False.
    force_sitelog : bool, optional
        If a single sitelog is provided, force sitelog-based header
        values when RINEX's header and sitelog site name do not correspond.
        If several sitelogs are provided, skip badly-formated sitelogs.
        The default is False.
    ignore : bool, optional
        Ignore firmware changes between instrumentation periods
        when getting header values info from sitelogs. The default is False.
    ninecharfile : str, optional
        Path of a file that contains 9-char. site names from the M3G database.
        The default is None.
    no_hatanaka : bool, optional
        Skip high-level RINEX-specific Hatanaka compression
        (performed per default).
        The default is False.
    compression : str, optional
        Set low-level RINEX file compression.
        acceptable values : 'gz' (recommended to fit IGS standards), 'Z', None.
        The default is 'gz'.
    relative : str, optional
        Reconstruct files relative subfolders.
        You have to indicate the common parent folder,
        that will be replaced with the output folder. The default is ''.
    verbose : bool, optional
        set the level of verbosity
        (False for the INFO level, True for the DEBUG level).
        The default is True.
    full_history : bool, optional
        Add the full history of the station in
        the RINEX's header as comment.
    filename_style : str, optional
        Set the RINEX filename style.
        acceptable values : 'basic' (per default), 'flex', 'exact'.
        * 'basic': a simple mode to apply a strict filename period (01H or 01D),
        being compatible with the IGS conventions.
        e.g.: `FNG000GLP_R_20242220000_01D_30S_MO.crx.gz`
        * 'flex': the filename period is tolerant and corresponds to
        the actual data content, but then can be odd (e.g. 07H, 14H...).
        The filename start time is rounded to the hour.
        e.g.: `FNG000GLP_R_20242221800_06H_30S_MO.crx.gz`
        * 'exact': the  filename start time is strictly the one of the
        first epoch in the RINEX.
        Useful for some specific cases needing splicing.
        e.g.: `FNG000GLP_R_20242221829_06H_30S_MO.crx.gz`
        The default is 'basic'.
    return_lists : dict, optional
        Specific option for file distribution through a GLASS node.
        Store the rinexmoded RINEXs in a dictionary
        to activates it, give a dict as input (an empty one - dict() works)
        The default is None.
    station_info: str, optional
        Path of a GAMIT station.info file to obtain GNSS site
        metadata information (needs also lfile_apriori option)
    lfile_apriori: str, optional
        Path of a GAMIT apriori apr/L-File to obtain GNSS site
        position and DOMES information (needs also station_info option)
    force_fake_coords: bool, optional
        When using GAMIT station.info metadata without apriori coordinates
        in the L-File, gives fake coordinates at (0°,0°) to the site
    remove: bool, optional
        Remove input RINEX file if the output RINEX is correctly written
        The default is False.
    keep_rnx_rec: bool, optional
        Keep the RINEX receiver header record in the output RINEX.
        Metadata from the external source (e.g. sitelogs) will not be modded.
        The default is False.
    round_instru_dates: bool
        If True, instrumentation dates (for receiver & antenna)
        in the metadata (sitelogs...) are rounded to day boundaries.
        The default is False.

    Raises
    ------
    RinexModInputArgsError
        Something is wrong with the input arguments.
    RinexFileError
        Something is wrong with the input RINEX File.

    Returns
    -------
    outputfile : str
        the path of the rinexmoded RINEX

    OR

    return_lists : dict
        a dictionary of rinexmoded RINEXs for GLASS distribution.
    """

    now = datetime.now().astimezone()

    if verbose:
        logger = rimo_log.logger_define("DEBUG", logfile=None, level_logfile="DEBUG")
    else:
        logger = rimo_log.logger_define("INFO", logfile=None, level_logfile="INFO")

    logger.info("# Inp. file: %s", rinexfile)

    if relative:
        if not relative in rinexfile:
            errmsg = "{:110s} - {}".format(
                "The relative subfolder can not be reconstructed for RINEX file",
                rinexfile,
            )
            logger.error(errmsg)
            raise rimo_cor.RinexModInputArgsError(errmsg)

        # We construct the output path with relative path between file name and parameter
        relpath = os.path.relpath(os.path.dirname(rinexfile), relative)
        myoutputfolder = os.path.join(outputfolder, relpath)
        if not os.path.isdir(myoutputfolder):
            os.makedirs(myoutputfolder)
    elif os.path.basename(outputfolder) == "INPUT_FOLDER":
        myoutputfolder = os.path.dirname(rinexfile)
    else:
        myoutputfolder = outputfolder

    if not modif_kw:
        modif_kw = dict()

    if (
        os.path.abspath(os.path.dirname(rinexfile)) == myoutputfolder
        and not outputfolder == "IDEM"
    ):
        errmsg = "{:110s} - {}".format(
            "Input and output folders are the same!", rinexfile
        )
        logger.error(errmsg)
        raise rimo_cor.RinexFileError(errmsg)

    if outputfolder == "IDEM":
        logger.warning("The output folder is forced as the same one as the input one")

    if not os.path.exists(outputfolder):
        logger.debug("The output folder does not exists, creating it")
        os.makedirs(outputfolder)

    if longname and shortname:
        errmsg = "longname and shortname are mutually exclusive"
        logger.error(errmsg)
        raise rimo_cor.RinexModInputArgsError(errmsg)

    ###########################################################################
    ########## Open the rinex file as an object
    rnxobj = rimo_rnx.RinexFile(rinexfile, force_rnx_load=force_rnx_load)

    if rnxobj.status:
        errmsg = "{:110s} - {}".format(rnxobj.status, rinexfile)
        logger.error(errmsg)
        raise rimo_cor.RinexFileError(errmsg)

    logger.debug("RINEX Origin Metadata :\n" + rnxobj.get_header()[0])

    # apply tolerant / strict (per default) file period
    if filename_style == "basic":
        rnxobj.mod_file_period_basic()

    # Check that the provided marker is a 4-char site name
    if marker and (len(marker) != 4 and len(marker) != 9):
        errmsg = "The site name provided is not 4 or 9-char valid: " + marker
        logger.error(errmsg)
        raise rimo_cor.RinexModInputArgsError(errmsg)

    # Get the 4 char > 9 char dictionnary from the input list
    nine_char_dict = dict()  # in any case, nine_char_dict is initialized
    if ninecharfile:
        if not os.path.isfile(ninecharfile):
            errmsg = "The specified 9-chars. list file does not exists: " + ninecharfile
            logger.error(errmsg)
            raise rimo_cor.RinexModInputArgsError(errmsg)

        nine_char_dict = rimo_cor.read_ninecharfile(ninecharfile)

    # set the marker as Rinex site, if any
    # This preliminary set_site is for th research of the right sitelog
    # a second set_site will take place a just after with more details
    if marker:
        # We store the old site name to add a comment in rinex file's header
        ## modif_marker = rnxobj.get_site(True,False) ### Useless...
        rnxobj.set_site(marker)

    ## warning if no metadata at all is not provided
    if not sitelog and not modif_kw and (not station_info or not lfile_apriori):
        logger.warning(
            "No sitelog nor keywords nor station.info+lfile provided. "
            "Per default rec.'s header will remain & no new "
            "metdata will be written!"
        )

    # initialize the metadata object list (to avoid linter warning)
    mdaobjs_lis = []
    ## sitelogs
    if sitelog:
        mdaobjs_lis = rimo_cor.metadata_input_manage(sitelog, force=force_sitelog)
    ## GAMIT files
    # We read the GAMIT files only if no input 'sitelog' variable is given.
    # Indeed, the GAMIT files might have been read outside this function
    # (most likely actually). The already read GAMIT files are then stored
    # in the 'sitelog' variable as a list of MetaData objects

    if (station_info and not lfile_apriori) or (not station_info and lfile_apriori):
        errmsg = "station_info and lfile_apriori must be provided together"
        logger.critical(errmsg)
        raise rimo_cor.RinexModInputArgsError(errmsg)

    ### load the metadata from sitelog or GAMIT files if any
    if (station_info and lfile_apriori) and not sitelog:
        mdaobjs_lis = rimo_cor.gamit2mda_objs(
            station_info, lfile_apriori, force_fake_coords=force_fake_coords
        )

    ### find the right MetaData object corresponding to the RINEX
    if sitelog or (station_info and lfile_apriori):
        mdaobj = rimo_cor.find_mda4site(rnxobj, mdaobjs_lis, force=force_sitelog)
        logger.debug("metadata used: %s", mdaobj)
    else:
        mdaobj = None

    ###########################################################################
    ########## Handle the similar options to set the site code
    ### Priority for the Country source
    # 0) if --country is given it overrides everything,
    #    but at the end of the tests
    # 1) the marker option if 9 char are given
    # 2) the nine_char_dict from the ninecharfile option
    # 3) the MetaData object (most useful actually),
    #    but we maintain a fallback mechanism here if the sitelog is wrong
    # 4) last chance: test if the country code we get from the input 9-char
    #    code is not XXX. If so, we keep it
    # Finally, set default value for the monument & country codes
    #
    rnx_4char = rnxobj.get_site(True, True)
    rnx_9char = rnxobj.get_site(False, False)

    if marker and len(marker) == 9:
        monum = marker[4:6]
        cntry = marker[6:]
    elif ninecharfile:
        if not rnx_4char in nine_char_dict:
            logger.warning(
                "Site's missing in the input 9-char. file: %s", rinexfile
            )
            monum = "00"
            cntry = "XXX"
        else:
            monum = nine_char_dict[rnx_4char].upper()[4:6]
            cntry = nine_char_dict[rnx_4char].upper()[6:]
    elif mdaobj:
        monum = "00"
        cntry = mdaobj.get_country()
    elif rnx_9char[6:] != "XXX":
        monum = rnx_9char[4:6]
        cntry = rnx_9char[6:]
    else:
        monum = "00"
        cntry = "XXX"

    if country:
        if len(country) == 3:
            cntry = country
        else:
            logger.warning(
                "Input country code is not 3 chars. RINEX will not be properly renamed: %s",
                rinexfile,
            )

    if cntry == "XXX":
        logger.warning(
            "Site's country not retrieved. RINEX will not be properly renamed: %s",
            rinexfile,
        )

    rnxobj.set_site(rnx_4char, monum, cntry)

    ###########################################################################
    ########## Remove previous comments
    if mdaobj or modif_kw or (station_info and lfile_apriori):
        rnxobj.clean_rinexmod_comments(clean_history=True)

    ###########################################################################
    ########## Apply the MetaData object on the RinexFile object
    if mdaobj:
        rnxobj = rimo_cor.apply_mda2rnxobj(
            rnxobj, mdaobj,
            ignore=ignore,
            keep_rnx_rec=keep_rnx_rec,
            round_instru_dates=round_instru_dates,
        )
        logger.debug("RINEX Sitelog-Modified Metadata :\n" + rnxobj.get_header()[0])
        modif_source_metadata = mdaobj.filename
    else:
        modif_source_metadata = ""

    ###########################################################################
    ########## Apply the modif_kw dictionnary on the RinexFile object
    if modif_kw:
        # Checking input keyword modification arguments
        rimo_cor.modif_kw_check(modif_kw)

        modif_source_kw = "keywords:" + " ".join(modif_kw.keys())
        rnxobj = rimo_cor.apply_modifkw2rnxobj(rnxobj, modif_kw)
        logger.debug(
            "RINEX Manual Keywords-Modified Metadata:\n" + rnxobj.get_header()[0]
        )
    else:
        modif_source_kw = ""

    ###########################################################################
    ########## Apply the site as the MARKER NAME within the RINEX
    # Must be after apply_mda2rnxobj and apply_modifkw2rnxobj
    # apply only is modif_kw does not overrides it (it is the overwhelming case)
    if "marker_name" not in modif_kw.keys():
        rnxobj.mod_marker(rnxobj.get_site(False, False, True))

    ###########################################################################
    ########## Correct the first and last time obs
    rnxobj.mod_time_obs(rnxobj.start_date, rnxobj.end_date)

    ###########################################################################
    ########## Add comment in the header
    githash = rimo_cor.get_git_hash() # git hash is desactivated
    vers_num = rimo.__version__ # + " " + githash[-3:]
    vers_num = vers_num.replace("beta", "b")
    rnxobj.add_prg_run_date_comment("RinexMod " + vers_num, "METADATA UPDATE")
    if githash != "xxxxxxx":
        rnxobj.add_comment("RinexMod Git commit hash: " + githash)
    rnxobj.add_comment("RinexMod / IPGP-OVS (github.com/IPGP/rinexmod)")
    rnxobj.add_comment(
        "rinexmoded on {}".format(datetime.strftime(now, "%Y-%m-%d %H:%M%z"))
    )
    if mdaobj:
        rnxobj.add_comment("rinexmoded with {}".format(modif_source_metadata))
    if modif_kw:
        rnxobj.add_comment("rinexmoded with {}".format(modif_source_kw))
    # if marker: ##### Useless...
    #    rnxobj.add_comment('filename assigned from {}'.format(modif_marker))

    ###########################################################################
    ########## Write the station history in the header
    if mdaobj and full_history:
        title = ["- - - - - - - - - - - -", "SITE FULL HISTORY"]
        rnxobj.add_comments(title + mdaobj.rinex_full_history_lines())

    ###########################################################################
    ########## Sort the header
    rnxobj.sort_header()

    ###########################################################################
    ########## we regenerate the filenames
    if shortname or (rnxobj.name_conv == "SHORT" and not longname):
        apply_longname = False
    elif longname or (rnxobj.name_conv == "LONG" and not shortname):
        apply_longname = True
    else:
        apply_longname = True

    if apply_longname:
        rnxobj.get_longname(
            inplace_set=True,
            compression="",
            filename_style=filename_style,
            data_source=rnxobj.data_source,
        )
    else:
        rnxobj.get_shortname(
            inplace_set=True, compression="", filename_style=filename_style
        )

    # NB: here the compression type must be forced to ''
    #     it will be added in the next step in write_to_path

    ###########################################################################
    ########## We convert the file back to Hatanaka Compressed Rinex

    # NB: this test is complcated, ambiguous and not very useful => disabled 2025-01-05

    # if apply_longname and not compression:
    #     # If not specified, we set compression to gz when file changed to longname
    #     output_compression = "gz"
    # elif not compression:
    #     output_compression = rnxobj.compression
    # else:
    #     output_compression = compression

    output_compression = compression

    ###########################################################################
    ########## Writing output file
    try:
        outputfile = rnxobj.write_to_path(
            myoutputfolder, compression=output_compression, no_hatanaka=no_hatanaka
        )
        logger.info("# Out. file: " + outputfile)
    except hatanaka.hatanaka.HatanakaException as e:
        errmsg ="{:110s} - {}".format("File could not be written - hatanaka exception", rinexfile)
        logger.error(errmsg)
        outputfile = None
        raise e

    if remove and os.path.isfile(outputfile):
        logger.info("Input file removed: %s", rinexfile)
        os.remove(rinexfile)

    ###########################################################################
    # Construct return dict by adding key if doesn't exists
    # and appending file to corresponding list
    if type(return_lists) is dict:
        return_lists = rimo_cor.rtun_lsts_make(rnxobj, return_lists)
        final_return = return_lists
    else:
        ###########################################################################
        # if no return dict given, return simply the path of the outputfile
        final_return = outputfile

    return final_return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 27/06/2025 15:10:30

@author: psakic
"""
import warnings
warnings.warn(
    f"from rinexmod import rinexmod_api is deprecated, use from import rinexmod.api as rimo_api instead.",
    category=DeprecationWarning,
    stacklevel=2,
)


# from rinexmod import rinexmod_api
# from rinexmod import api as rimo_api

from .api import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
This script will get last version of sitelogs from M3G repository and write them
in an observatory dependent subfolder set in 'observatories'.
The -d --delete option will delete old version as to get only last version even
in a name changing case.
The -o --obsevatory option will download sitelogs for the specified observatory
only. Possible values : OVSM|OVSG|OVPF|REVOSIMA
USE :


* OUTPUTFOLDER : Folder where to write the downloaded sitelogs.

OPTION :

* -d : delete : Delete old sitelogs in storage folder. This allows to have only
                the last version, as version changing sitelogs changes of name.

* -m : move : Move old sitelogs into the given archive folder

* -o : observatory : IPGP's observatory ID that will be used to filter
                     sitelogs to download. Valid values are : OVSM|OVSG|OVPF|REVOSIMA.

* -r : root folder : per default, an observatory-specific folder is created
                     to store the corresponding sitelogs.
                     This option stores them in OUTPUTFOLDER's root

* -s : svn : a mode to maintain the legacy OVS's SVN folders.
             download the sitelog of a single obs and perform a 'svn commit'
             a single observatory must be given with -o option.
             the root folder option is automatically activated (-r)

* -f : force : Force download even if an identical sitelog already exists locally

* -e : exclude : Site(s) you want to exclude from download.
                 Provide as input 4 or 9 character site codes separated with spaces

EXAMPLE:

./get_m3g_sitelogs.py OUTPUTFOLDER (-d) (-o OVSM|OVSG|OVPF|REVOSIMA) (-r) (-s)

2021-10-18 FL
2023-03-21 PS
"""

import requests
import subprocess
import hashlib
import os, glob
import shutil
import sys

def get_m3g_sitelogs(
    sitelogsfolder,
    delete=False,
    observatory=None,
    root_folder=False,
    svn_mode=False,
    move_folder=None,
    force=False,
    exclude=[],
    geodesyml=False,
):
    """
    Downloads the latest version of sitelogs from the M3G repository and writes them
    to an observatory-dependent subfolder set in 'observatories'.

    Parameters
    ----------
    sitelogsfolder : str
        The folder where the downloaded sitelogs will be stored.
    delete : bool
        If True, deletes old sitelogs in the storage folder to keep only the latest version.
    observatory : str or list of str, optional
        IPGP's observatories IDs to filter sitelogs to download.
        Valid values are: OVSM, OVSG, OVPF, REVOSIMA, OGA.
        Default is None.
    root_folder : bool, optional
        If True, stores sitelogs in the root of OUTPUTFOLDER. Default is False.
    svn_mode : bool, optional
        If True, maintains the legacy OVS's SVN folders,
        downloads the sitelog of a single observatory, and performs an 'svn commit'.
        Default is False.
    move_folder : str, optional
        The folder where old sitelogs will be moved. Default is None.
    force : bool, optional
        If True, forces download even if an identical sitelog already exists locally.
        Default is False.
    exclude : list, optional
        List of site codes (4 or 9 characters) to exclude from download.
        Default is an empty list.
    geodesyml : bool, optional
        If True, downloads GeodesyML files instead of sitelogs.

    Returns
    -------
    None
    """
    # the root folder option is automatically activated for the SVN mode
    if svn_mode:
        root_folder = True

    # a single observatory must be given for the SVN mode
    if svn_mode and not observatory:
        sys.exit(
            "when using SVN mode (-s), a single observatory must be given with -o option"
        )

    if not isinstance(observatory,list):
        observatory = [observatory]

    # M3G country webservice. We use countries to filter IPGP's obs stations.
    country_url = "https://gnss-metadata.eu/v1/sitelog/metadata-list?country="
    country_url = "https://gnss-metadata.eu/v1/sitelog/metadata-list?network="

    # # Observatories. First field is the counrty code in M3G's webservice.
    # # Second field is the name of the observatory and is used to build the subfolder
    # observatories = {
    #     "MTQ": "OVSM",
    #     "GLP": "OVSG",
    #     "BLM": "OVSG",  ### St Barthelemy
    #     "MAF": "OVSG",  ### St Martin (no station there for the moment)
    #     "REU": "OVPF",
    #     "MYT": "REVOSIMA",
    #     "ATF": "REVOSIMA",
    #     "DJI": "OGA",
    # }  ### TAAF aka Terres Australes

    # (No more country but network 2025-03)
    net_obs_dic_all = {
        "MQ": "OVSM",
        "GL": "OVSG",
        "PF": "OVPF",
        "QM": "REVOSIMA",
        "DJ": "OGA",
    }

    # If observatory names are given, only download its/their sitelogs
    if observatory and not (len(observatory) == 1 and not observatory[0]):
        net_obs_dic = {
            net_key: obs_val
            for net_key, obs_val in net_obs_dic_all.items()
            if obs_val in observatory
        }
    else:
        net_obs_dic = net_obs_dic_all

    ## complex for and ifs structure (2025-04)
    # if not (len(observatory) == 1 and not observatory[0]):
    #     for obs in observatory:
    #         if obs in net_obs_dic.values():
    #             net_obs_dic_cln = dict()
    #             for net_key, obs_val in net_obs_dic.items():
    #                 if obs_val in obs:
    #                     net_obs_dic_cln[net_key] = obs_val
    #
    #             net_obs_dic = net_obs_dic_cln

    ## simple way to filter observatories if observatories is bijective
    ## not the case anymore because ATF, BLM ... (2022-10)
    # net_key = list(observatories.keys())[list(observatories.values()).index(obs)]
    # observatories = {net_key: observatories[net_key]}

    # Check that output folder exists
    if not os.path.exists(sitelogsfolder):
        print("# ERROR : The output folder for sitelogs doesn't exist")
        return

    if move_folder and not os.path.exists(move_folder):
        print("# ERROR : The archive folder for old sitelogs doesn't exist")
        return

    # Check that obs' subfolder exists. If not, creates.
    for obs in [*net_obs_dic.values()]:
        if not root_folder:
            obs_path = os.path.join(sitelogsfolder, obs)
        else:
            obs_path = sitelogsfolder

        if not os.path.exists(obs_path):
            os.mkdir(obs_path)

    file_local_paths = []

    for net in net_obs_dic.keys():

        obs_url = country_url + net

        obs_infos = requests.get(obs_url)
        obs_infos = obs_infos.content.decode("utf-8")
        obs_infos = obs_infos.splitlines()
        obs_infos = obs_infos[1:]  # remove header
        obs_infos = list(reversed(obs_infos))  # list is reversed per def.

        if not root_folder:
            obs_path = os.path.join(sitelogsfolder, net_obs_dic[net])
        else:
            obs_path = sitelogsfolder

        print(
            "###### Downloading {} ({}) sitelogs from M3G to {}".format(
                net_obs_dic[net], net, obs_path
            )
        )

        for line in obs_infos:
            line = line.split()

            # station_name = line[0]

            # If station is declared in M3G but sitelog not available yet
            if len(line) < 6:
                print("### " + line[0] + " : not available ###")
                continue

            sitelog_md5 = line[1]
            sitelog_url = line[5]
            sitelog_name = line[2]
            geodesyml_url = line[6]
            geodesyml_name = line[3]

            if geodesyml:
                file_name = geodesyml_name
                file_url = geodesyml_url
            else:
                file_name = sitelog_name
                file_url = sitelog_url

            file_local_path = os.path.join(obs_path, file_name)
            file_local_paths.append(file_local_path)

            ### skip excluded stations
            exclude = [e.lower() for e in exclude]
            if file_name[:4] in exclude or file_name[:9] in exclude:
                print("### " + file_name + " skip (excluded) ###")
                continue

            ### get the checksum for the existing sitelog, if any
            if os.path.exists(file_local_path):
                local_file = open(file_local_path, "rb")
                local_md5 = hashlib.md5(local_file.read()).hexdigest()
            else:
                local_md5 = None

            if force or (local_md5 != sitelog_md5):
                print("### " + file_name + " download ###")
                ##Dowload the sitelog
                r = requests.get(file_url, allow_redirects=True)
                # print(r.status_code)
                content = r.content  # .rstrip()
                open(file_local_path, "wb").write(content)
                # subprocess.call(['wget',
                #                  '--no-check-certificate',
                #                  file_url,'-q',
                #                  '-O', file_local_path])

            else:
                print("### " + file_name + " skip (already exists) ###")

            ### get existing old sitelogs for moving or delete
            if move_folder or delete:
                old_sitelogs_mv = glob.glob(f"{obs_path}/*{file_name[:9]}*.log")
                for f in old_sitelogs_mv:
                    if f == file_local_path:
                        # f is the new sitelog
                        continue
                    elif move_folder:
                        print(f"### {os.path.basename(f)} moved to archive folder ###")
                        shutil.move(f, move_folder)
                    elif delete:
                        print(f"### {os.path.basename(f)} deleted ###")
                        os.remove(f)

    if svn_mode:
        print("### SVN add/commit of the downloaded sitelogs")
        for sitelog_local in file_local_paths:
            subprocess.call(["svn", "add", sitelog_local])
        subprocess.call(
            ["svn", "commit", "-m", "get_m3g_sitelogs auto commit", sitelogsfolder]
        )

    # # Other tecnhique using node webservice that will send back less information
    # # about the sitelogs but will filter with the node ID, and will return all
    # # sitelogs belonging to IPGP node
    #
    # node_url = "https://gnss-metadata.eu/v1/node/view?id="
    # node_id = '5e4d07d9468524145a7cf0f2' # IPGP
    #
    # file_url = "https://gnss-metadata.eu/v1/sitelog/exportlog?id="
    #
    # # Getting station list related to IPGP node
    # node_url = node_url + node_id
    #
    # node_infos = requests.get(node_url)
    # node_infos = node_infos.content.decode('utf-8')
    # node_infos = json.loads(node_infos)
    #
    # station_list = node_infos['stations']
    #
    # for station in station_list:
    #
    #     output = os.path.join(sitelogsfolder, observatories[station[-3:]], station.lower() + '.log')


def main():
    import argparse

    # Parsing Args
    parser = argparse.ArgumentParser(
        description="Get the last versions of the IPGP sitelogs sorted on the M3G Database"
    )
    parser.add_argument(
        "sitelogsfolder",
        type=str,
        help="Output folder where to store downloaded sitelogs",
    )
    parser.add_argument(
        "-d",
        "--delete",
        help="Delete old sitelogs in the output folder. "
             "This allows to have only the last version, "
             "as version changing sitelogs changes of name.",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--move",
        help="Move old sitelogs into the given archive folder",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-o",
        "--observatory",
        nargs="+",
        help="Download sitelogs for some specific IPGP's observatories."
             "Valid values are : OVSM OVSG OVPF REVOSIMA OGA",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-r",
        "--root",
        help="Store the sitelogs in OUTPUTFOLDER root."
             "(per default, an observatory-specific folder is"
             "created to store the corresponding sitelogs.)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-s",
        "--svn",
        help="A mode to maintain the legacy OVS SVN folder."
             "Download the sitelog of a single obs and perform a svn commit."
             "A single observatory must be given with -o option."
             "The root folder option is automatically activated (-r)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-f",
        "--force",
        help="Force download even if an identical sitelog already exists locally",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-e",
        "--exclude",
        help="Site(s) you want to exclude from download. "
             "Provide as input 4 or 9 character site codes separated with spaces",
        nargs="+",
        default=[],
    )

    parser.add_argument(
        "-g",
        "--geodesyml",
        help="Download GeodesyML files instead of sitelogs",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()

    get_m3g_sitelogs(
        sitelogsfolder=args.sitelogsfolder,
        delete=args.delete,
        observatory=args.observatory,
        root_folder=args.root,
        svn_mode=args.svn,
        move_folder=args.move,
        force=args.force,
        exclude=args.exclude,
        geodesyml=args.geodesyml,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Extract metadata from crz file.

With -p option, will plot the file's samples intervals

EXAMPLE:
./crzmeta.py  RINEXFILE (-p)

REQUIREMENTS :

You need Python Hatanaka library from Martin Valgur:

pip install hatanaka

2019-12-13 Félix Léger - leger@ipgp.fr
"""

from rinexmod.rinexfile import RinexFile


def crzmeta(rinexfile, plot):

    # Declare the rinex file as an object
    rinexfileobj = RinexFile(rinexfile)

    if rinexfileobj.status == 1:
        print('{:45s} - {}'.format('01 - The specified file does not exists', rinexfile))
        return

    if rinexfileobj.status == 2:
        print('{:45s} - {}'.format('02 - Not an observation Rinex file', rinexfile))
        return

    if rinexfileobj.status == 3:
        print('{:45s} - {}'.format('03 - Invalid or empty Zip file', rinexfile))
        return

    if rinexfileobj.status == 4:
        print('{:45s} - {}'.format('04 - Invalid Compressed Rinex file', rinexfile))
        return

    if rinexfileobj.status == 5:
        print('{:45s} - {}'.format('05 -Less than two epochs in the file', rinexfile))
        return

    print(rinexfileobj.get_header()[0])

    # We reload _get_sample_rate method with plot set to true
    if plot:
        rinexfileobj.get_sample_rate(plot = True)

    return


def main():
    import argparse

    # Parsing Args
    parser = argparse.ArgumentParser(description='Read a Sitelog file and create a CSV file output')
    parser.add_argument('rinexfile', type=str, help='Input rinex list file to process')
    parser.add_argument('-p', '--plot', help='Plot file\'s samples intervals', action='store_true', default=0)
    args = parser.parse_args()

    rinexfile = args.rinexfile
    plot = args.plot

    crzmeta(rinexfile, plot)

if __name__ == '__main__':
    main()
